# sage_setup: distribution = sagemath-categories
r"""
Hypergeometric functions

This module implements manipulation of infinite hypergeometric series
represented in standard parametric form (as `\,_pF_q` functions).

AUTHORS:

- Fredrik Johansson (2010): initial version

- Eviatar Bach (2013): major changes

EXAMPLES:

Examples from :issue:`9908`::

    sage: # needs sage.symbolic
    sage: maxima('integrate(bessel_j(2, x), x)').sage()
    1/24*x^3*hypergeometric((3/2,), (5/2, 3), -1/4*x^2)
    sage: sum(((2*I)^x/(x^3 + 1)*(1/4)^x), x, 0, oo)
    hypergeometric((1, 1, -1/2*I*sqrt(3) - 1/2, 1/2*I*sqrt(3) - 1/2),...
    (2, -1/2*I*sqrt(3) + 1/2, 1/2*I*sqrt(3) + 1/2), 1/2*I)
    sage: res = sum((-1)^x/((2*x + 1)*factorial(2*x + 1)), x, 0, oo)
    sage: res                                   # not tested (depends on maxima version)
    hypergeometric((1/2,), (3/2, 3/2), -1/4)
    sage: res in [hypergeometric((1/2,), (3/2, 3/2), -1/4), sin_integral(1)]
    True

Simplification (note that ``simplify_full`` does not yet call
``simplify_hypergeometric``)::

    sage: # needs sage.symbolic
    sage: hypergeometric([-2], [], x).simplify_hypergeometric()
    x^2 - 2*x + 1
    sage: hypergeometric([], [], x).simplify_hypergeometric()
    e^x
    sage: a = hypergeometric((hypergeometric((), (), x),), (),
    ....:                    hypergeometric((), (), x))
    sage: a.simplify_hypergeometric()
    1/((-e^x + 1)^e^x)
    sage: a.simplify_hypergeometric(algorithm='sage')
    1/((-e^x + 1)^e^x)

Equality testing::

    sage: bool(hypergeometric([], [], x).derivative(x) ==                               # needs sage.symbolic
    ....:      hypergeometric([], [], x))  # diff(e^x, x) == e^x
    True
    sage: bool(hypergeometric([], [], x) == hypergeometric([], [1], x))                 # needs sage.symbolic
    False

Computing terms and series::

    sage: # needs sage.symbolic
    sage: var('z')
    z
    sage: hypergeometric([], [], z).series(z, 0)
    Order(1)
    sage: hypergeometric([], [], z).series(z, 1)
    1 + Order(z)
    sage: hypergeometric([], [], z).series(z, 2)
    1 + 1*z + Order(z^2)
    sage: hypergeometric([], [], z).series(z, 3)
    1 + 1*z + 1/2*z^2 + Order(z^3)

    sage: # needs sage.symbolic
    sage: hypergeometric([-2], [], z).series(z, 3)
    1 + (-2)*z + 1*z^2
    sage: hypergeometric([-2], [], z).series(z, 6)
    1 + (-2)*z + 1*z^2
    sage: hypergeometric([-2], [], z).series(z, 6).is_terminating_series()
    True
    sage: hypergeometric([-2], [], z).series(z, 2)
    1 + (-2)*z + Order(z^2)
    sage: hypergeometric([-2], [], z).series(z, 2).is_terminating_series()
    False

    sage: hypergeometric([1], [], z).series(z, 6)                                       # needs sage.symbolic
    1 + 1*z + 1*z^2 + 1*z^3 + 1*z^4 + 1*z^5 + Order(z^6)
    sage: hypergeometric([], [1/2], -z^2/4).series(z, 11)                               # needs sage.symbolic
    1 + (-1/2)*z^2 + 1/24*z^4 + (-1/720)*z^6 + 1/40320*z^8 +...
    (-1/3628800)*z^10 + Order(z^11)

    sage: hypergeometric([1], [5], x).series(x, 5)                                      # needs sage.symbolic
    1 + 1/5*x + 1/30*x^2 + 1/210*x^3 + 1/1680*x^4 + Order(x^5)

    sage: sum(hypergeometric([1, 2], [3], 1/3).terms(6)).n()                            # needs sage.symbolic
    1.29788359788360
    sage: hypergeometric([1, 2], [3], 1/3).n()                                          # needs sage.symbolic
    1.29837194594696
    sage: hypergeometric([], [], x).series(x, 20)(x=1).n() == e.n()                     # needs sage.symbolic
    True

Plotting::

    sage: # needs sage.symbolic
    sage: f(x) = hypergeometric([1, 1], [3, 3, 3], x)
    sage: plot(f, x, -30, 30)                                                           # needs sage.plot
    Graphics object consisting of 1 graphics primitive
    sage: g(x) = hypergeometric([x], [], 2)
    sage: complex_plot(g, (-1, 1), (-1, 1))                                             # needs sage.plot
    Graphics object consisting of 1 graphics primitive

Numeric evaluation::

    sage: # needs sage.symbolic
    sage: hypergeometric([1], [], 1/10).n()  # geometric series
    1.11111111111111
    sage: hypergeometric([], [], 1).n()  # e
    2.71828182845905
    sage: hypergeometric([], [], 3., hold=True)
    hypergeometric((), (), 3.00000000000000)
    sage: hypergeometric([1, 2, 3], [4, 5, 6], 1/2).n()
    1.02573619590134
    sage: hypergeometric([1, 2, 3], [4, 5, 6], 1/2).n(digits=30)
    1.02573619590133865036584139535
    sage: hypergeometric([5 - 3*I], [3/2, 2 + I, sqrt(2)], 4 + I).n()
    5.52605111678803 - 7.86331357527540*I
    sage: hypergeometric((10, 10), (50,), 2.)
    -1705.75733163554 - 356.749986056024*I

Conversions::

    sage: maxima(hypergeometric([1, 1, 1], [3, 3, 3], x))                               # needs sage.symbolic
    hypergeometric([1,1,1],[3,3,3],_SAGE_VAR_x)
    sage: hypergeometric((5,), (4,), 3)._sympy_()                                   # needs sympy sage.symbolic
    hyper((5,), (4,), 3)
    sage: hypergeometric((5, 4), (4, 4), 3)._mathematica_init_()                        # needs sage.symbolic
    'HypergeometricPFQ[{5,4},{4,4},3]'

Arbitrary level of nesting for conversions::

    sage: maxima(nest(lambda y: hypergeometric([y], [], x), 3, 1))                      # needs sage.symbolic
    1/(1-_SAGE_VAR_x)^(1/(1-_SAGE_VAR_x)^(1/(1-_SAGE_VAR_x)))
    sage: maxima(nest(lambda y: hypergeometric([y], [3], x), 3, 1))._sage_()            # needs sage.symbolic
    hypergeometric((hypergeometric((hypergeometric((1,), (3,), x),), (3,),...
    x),), (3,), x)
    sage: nest(lambda y: hypergeometric([y], [], x), 3, 1)._mathematica_init_()         # needs sage.symbolic
    'HypergeometricPFQ[{HypergeometricPFQ[{HypergeometricPFQ[{1},{},x]},...

The confluent hypergeometric functions can arise as solutions to second-order
differential equations (example from `here <http://ask.sagemath.org/question/
1168/how-can-one-use-maxima-kummer-confluent-functions>`_)::

    sage: var('m')                                                                      # needs sage.symbolic
    m
    sage: y = function('y')(x)                                                          # needs sage.symbolic
    sage: desolve(diff(y, x, 2) + 2*x*diff(y, x) - 4*m*y, y,                            # needs sage.symbolic
    ....:         contrib_ode=true, ivar=x)
    [y(x) == _K1*hypergeometric_M(-m, 1/2, -x^2) +...
     _K2*hypergeometric_U(-m, 1/2, -x^2)]

Series expansions of confluent hypergeometric functions::

    sage: hypergeometric_M(2, 2, x).series(x, 3)                                        # needs sage.symbolic
    1 + 1*x + 1/2*x^2 + Order(x^3)
    sage: hypergeometric_U(2, 2, x).series(x == 3, 100).subs(x=1).n()                   # needs sage.symbolic
    0.403652637676806
    sage: hypergeometric_U(2, 2, 1).n()                                                 # needs mpmath sage.symbolic
    0.403652637676806
"""

# ****************************************************************************
#       Copyright (C) 2010 Fredrik Johansson <fredrik.johansson@gmail.com>
#       Copyright (C) 2013 Eviatar Bach <eviatarbach@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from functools import reduce

from sage.arith.misc import binomial, factorial, rising_factorial
from sage.calculus.functional import derivative
from sage.functions.error import erf
from sage.functions.gamma import gamma
from sage.functions.hyperbolic import cosh, sinh
from sage.functions.log import exp, log
from sage.functions.other import sqrt, real_part
from sage.misc.lazy_import import lazy_import
from sage.misc.misc_c import prod
from sage.rings.infinity import Infinity
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.structure.element import Expression, get_coercion_model
from sage.symbolic.function import BuiltinFunction

lazy_import('sage.misc.latex', 'latex')

lazy_import('sage.symbolic.constants', 'pi')
lazy_import('sage.symbolic.ring', 'SR')

lazy_import('sage.libs.mpmath.utils', 'call', as_='_mpmath_utils_call')
lazy_import('mpmath', 'hyp1f1', as_='_mpmath_hyp1f1')
lazy_import('mpmath', 'hyper', as_='_mpmath_hyper')
lazy_import('mpmath', 'hyperu', as_='_mpmath_hyperu')


def rational_param_as_tuple(x):
    r"""
    Utility function for converting rational `\,_pF_q` parameters to
    tuples (which mpmath handles more efficiently).

    EXAMPLES::

        sage: from sage.functions.hypergeometric import rational_param_as_tuple
        sage: rational_param_as_tuple(1/2)
        (1, 2)
        sage: rational_param_as_tuple(3)
        3
        sage: rational_param_as_tuple(pi)                                               # needs sage.symbolic
        pi
    """
    try:
        x = x.pyobject()
    except AttributeError:
        pass
    try:
        if x.parent() is QQ:
            p = int(x.numer())
            q = int(x.denom())
            return p, q
    except AttributeError:
        pass
    return x


class Hypergeometric(BuiltinFunction):
    r"""
    Represent a (formal) generalized infinite hypergeometric series.

    It is defined as

    .. MATH::

        \,_pF_q(a_1, \ldots, a_p; b_1, \ldots, b_q; z)
        = \sum_{n=0}^{\infty} \frac{(a_1)_n \cdots (a_p)_n}{(b_1)_n
        \cdots(b_q)_n} \, \frac{z^n}{n!},

    where `(x)_n` is the rising factorial.
    """
    def __init__(self):
        """
        Initialize class.

        EXAMPLES::

            sage: maxima(hypergeometric)                                                # needs sage.symbolic
            hypergeometric

        TESTS::

            sage: F = hypergeometric([-4,2],[1],1)  # optional - maple
            sage: G = maple(F); G                   # optional - maple
            hypergeom([-4, 2],[1],1)
            sage: G.simplify()                      # optional - maple
            0
        """
        BuiltinFunction.__init__(self, 'hypergeometric', nargs=3,
                                 conversions={'mathematica':
                                              'HypergeometricPFQ',
                                              'maxima': 'hypergeometric',
                                              'maple': 'hypergeom',
                                              'sympy': 'hyper',
                                              'fricas': 'hypergeometricF'})

    def __call__(self, a, b, z, **kwargs):
        """
        Return symbolic hypergeometric function expression.

        INPUT:

        - ``a`` -- list or tuple of parameters
        - ``b`` -- list or tuple of parameters
        - ``z`` -- number or symbolic expression

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: hypergeometric([], [], 1)
            hypergeometric((), (), 1)
            sage: hypergeometric([], [1], 1)
            hypergeometric((), (1,), 1)
            sage: hypergeometric([2, 3], [1], 1)
            hypergeometric((2, 3), (1,), 1)
            sage: hypergeometric([], [], x)
            hypergeometric((), (), x)
            sage: hypergeometric([x], [], x^2)
            hypergeometric((x,), (), x^2)

        The only simplification that is done automatically is returning 1
        if ``z`` is 0. For other simplifications use the
        ``simplify_hypergeometric`` method.

        TESTS::

            sage: hypergeometric([2, 3, 4], [4, 1], 1)                                  # needs sage.symbolic
            hypergeometric((2, 3, 4), (4, 1), 1)
        """
        return BuiltinFunction.__call__(self,
                                        SR._force_pyobject(a),
                                        SR._force_pyobject(b),
                                        z, **kwargs)

    def _print_latex_(self, a, b, z):
        r"""
        TESTS::

            sage: latex(hypergeometric([1, 1], [2], -1))                                # needs sage.symbolic
            \,_2F_1\left(\begin{matrix} 1,1 \\ 2 \end{matrix} ; -1 \right)
        """
        aa = ",".join(latex(c) for c in a)
        bb = ",".join(latex(c) for c in b)
        z = latex(z)
        return (r"\,_{}F_{}\left(\begin{{matrix}} {} \\ {} \end{{matrix}} ; "
                r"{} \right)").format(len(a), len(b), aa, bb, z)

    def _eval_(self, a, b, z, **kwargs):
        """
        EXAMPLES::

            sage: hypergeometric([], [], 0)                                             # needs sage.symbolic
            1
        """
        if not isinstance(a, tuple) or not isinstance(b, tuple):
            raise TypeError("The first two parameters must be of type list")

        if not isinstance(z, Expression) and z == 0:  # Expression is excluded
            return Integer(1)                         # to avoid call to Maxima

    def _evalf_try_(self, a, b, z):
        """
        Call :meth:`_evalf_` if one of the arguments is numerical and none
        of the arguments are symbolic.

        OUTPUT:

        - ``None`` if we didn't succeed to call :meth:`_evalf_` or if
          the input wasn't suitable for it.

        - otherwise, a numerical value for the function.

        EXAMPLES::

            sage: hypergeometric._evalf_try_((1.0,), (2.0,), 3.0)                       # needs mpmath
            6.36184564106256
            sage: hypergeometric._evalf_try_((1.0, 1), (), 3.0)                         # needs mpmath
            -0.0377593153441588 + 0.750349833788561*I
            sage: hypergeometric._evalf_try_((1, 1), (), 3)    # exact input
            sage: hypergeometric._evalf_try_((x,), (), 1.0)    # symbolic               # needs sage.symbolic
            sage: hypergeometric._evalf_try_(1.0, 2.0, 3.0)    # not tuples
        """
        # We need to override this for hypergeometric functions since
        # the first 2 arguments are tuples and the generic _evalf_try_
        # cannot handle that.
        if not isinstance(a, tuple) or not isinstance(b, tuple):
            return None

        args = list(a) + list(b) + [z]
        if any(self._is_numerical(x) for x in args):
            if not any(isinstance(x, Expression) for x in args):
                p = get_coercion_model().common_parent(*args)
                return self._evalf_(a, b, z, parent=p)

    def _evalf_(self, a, b, z, parent=None, algorithm=None):
        """
        TESTS::

            sage: hypergeometric([1, 1], [2], -1).n()                                   # needs sage.symbolic
            0.693147180559945
            sage: hypergeometric([], [], RealField(100)(1))                             # needs sage.rings.real_mpfr sage.symbolic
            2.7182818284590452353602874714
        """
        if not isinstance(a, tuple) or not isinstance(b, tuple):
            raise TypeError("The first two parameters must be of type list")
        aa = [rational_param_as_tuple(c) for c in a]
        bb = [rational_param_as_tuple(c) for c in b]
        return _mpmath_utils_call(_mpmath_hyper, aa, bb, z, parent=parent)

    def _tderivative_(self, a, b, z, *args, **kwargs):
        """
        EXAMPLES::

            sage: hypergeometric([1/3, 2/3], [5], x^2).diff(x)                          # needs sage.symbolic
            4/45*x*hypergeometric((4/3, 5/3), (6,), x^2)
            sage: hypergeometric([1, 2], [x], 2).diff(x)                                # needs sage.symbolic
            Traceback (most recent call last):
            ...
            NotImplementedError: derivative of hypergeometric function with...
            respect to parameters. Try calling .simplify_hypergeometric()...
            first.
            sage: hypergeometric([1/3, 2/3], [5], 2).diff(x)                            # needs sage.symbolic
            0
        """
        diff_param = kwargs['diff_param']
        if diff_param in hypergeometric(a, b, 1).variables():  # ignore z
            raise NotImplementedError("derivative of hypergeometric function "
                                      "with respect to parameters. Try calling"
                                      " .simplify_hypergeometric() first.")
        t = (reduce(lambda x, y: x * y, a, 1) *
             reduce(lambda x, y: x / y, b, Integer(1)))
        return (t * derivative(z, diff_param) *
                hypergeometric([c + 1 for c in a], [c + 1 for c in b], z))

    class EvaluationMethods:

        def _fast_callable_(self, a, b, z, etb):
            """
            Override the ``fast_callable`` method.

            OUTPUT:

            A :class:`~sage.ext.fast_callable.ExpressionCall` representing the
            hypergeometric function in the expression tree.

            EXAMPLES::

                sage: # needs sage.symbolic
                sage: h = hypergeometric([], [], x)
                sage: from sage.ext.fast_callable import ExpressionTreeBuilder
                sage: etb = ExpressionTreeBuilder(vars=['x'])
                sage: h._fast_callable_(etb)
                {hypergeometric((), (), x)}(v_0)

                sage: var('x, y')                                                       # needs sage.symbolic
                (x, y)
                sage: f = fast_callable(hypergeometric([y], [], x), vars=[x, y])        # needs sage.symbolic
                sage: f(3, 4)                                                           # needs sage.symbolic
                hypergeometric((4,), (), 3)
            """
            return etb.call(self, *map(etb.var, etb._vars))

        def sorted_parameters(self, a, b, z):
            """
            Return with parameters sorted in a canonical order.

            EXAMPLES::

                sage: hypergeometric([2, 1, 3], [5, 4],                                 # needs sage.symbolic
                ....:                1/2).sorted_parameters()
                hypergeometric((1, 2, 3), (4, 5), 1/2)
            """
            return hypergeometric(sorted(a), sorted(b), z)

        def eliminate_parameters(self, a, b, z):
            """
            Eliminate repeated parameters by pairwise cancellation of identical
            terms in ``a`` and ``b``.

            EXAMPLES::

                sage: hypergeometric([1, 1, 2, 5], [5, 1, 4],                           # needs sage.symbolic
                ....:                1/2).eliminate_parameters()
                hypergeometric((1, 2), (4,), 1/2)
                sage: hypergeometric([x], [x], x).eliminate_parameters()                # needs sage.symbolic
                hypergeometric((), (), x)
                sage: hypergeometric((5, 4), (4, 4), 3).eliminate_parameters()          # needs sage.symbolic
                hypergeometric((5,), (4,), 3)
            """
            aa = list(a)  # tuples are immutable
            bb = list(b)
            q = qq = len(bb)
            i = 0
            while i < qq and aa:
                bbi = bb[i]
                if bbi in aa:
                    aa.remove(bbi)
                    bb.remove(bbi)
                    qq -= 1
                else:
                    i += 1
            if qq != q:
                return hypergeometric(aa, bb, z)
            return self

        def is_termwise_finite(self, a, b, z) -> bool:
            """
            Determine whether all terms of ``self`` are finite.

            Any infinite terms or ambiguous terms beyond the first
            zero, if one exists, are ignored.

            Ambiguous cases (where a term is the product of both zero
            and an infinity) are not considered finite.

            EXAMPLES::

                sage: # needs sage.symbolic
                sage: hypergeometric([2], [3, 4], 5).is_termwise_finite()
                True
                sage: hypergeometric([2], [-3, 4], 5).is_termwise_finite()
                False
                sage: hypergeometric([-2], [-3, 4], 5).is_termwise_finite()
                True
                sage: hypergeometric([-3], [-3, 4],
                ....:                5).is_termwise_finite()  # ambiguous
                False

                sage: # needs sage.symbolic
                sage: hypergeometric([0], [-1], 5).is_termwise_finite()
                True
                sage: hypergeometric([0], [0],
                ....:                5).is_termwise_finite()  # ambiguous
                False
                sage: hypergeometric([1], [2], Infinity).is_termwise_finite()
                False
                sage: (hypergeometric([0], [0], Infinity)
                ....:  .is_termwise_finite())  # ambiguous
                False
                sage: (hypergeometric([0], [], Infinity)
                ....:  .is_termwise_finite())  # ambiguous
                False
            """
            if z == 0:
                return 0 not in b
            if abs(z) == Infinity:
                return False
            for bb in b:
                if bb in ZZ and bb <= 0:
                    if any((aa in ZZ) and (bb < aa <= 0) for aa in a):
                        continue
                    return False
            return True

        def is_terminating(self, a, b, z):
            r"""
            Determine whether the series represented by ``self`` terminates
            after a finite number of terms.

            This happens if any of the
            numerator parameters are nonnegative integers (with no
            preceding nonnegative denominator parameters), or `z = 0`.

            If terminating, the series represents a polynomial of `z`.

            EXAMPLES::

                sage: hypergeometric([1, 2], [3, 4], x).is_terminating()                # needs sage.symbolic
                False
                sage: hypergeometric([1, -2], [3, 4], x).is_terminating()               # needs sage.symbolic
                True
                sage: hypergeometric([1, -2], [], x).is_terminating()                   # needs sage.symbolic
                True
            """
            if z == 0:
                return True
            for aa in a:
                if (aa in ZZ) and (aa <= 0):
                    return self.is_termwise_finite()
            return False

        def is_absolutely_convergent(self, a, b, z):
            r"""
            Determine whether ``self`` converges absolutely as an infinite
            series. ``False`` is returned if not all terms are finite.

            EXAMPLES:

            Degree giving infinite radius of convergence::

                sage: hypergeometric([2, 3], [4, 5],                                    # needs sage.symbolic
                ....:                6).is_absolutely_convergent()
                True
                sage: hypergeometric([2, 3], [-4, 5],                                   # needs sage.symbolic
                ....:                6).is_absolutely_convergent()  # undefined
                False
                sage: (hypergeometric([2, 3], [-4, 5], Infinity)                        # needs sage.symbolic
                ....:  .is_absolutely_convergent())  # undefined
                False

            Ordinary geometric series (unit radius of convergence)::

                sage: # needs sage.symbolic
                sage: hypergeometric([1], [], 1/2).is_absolutely_convergent()
                True
                sage: hypergeometric([1], [], 2).is_absolutely_convergent()
                False
                sage: hypergeometric([1], [], 1).is_absolutely_convergent()
                False
                sage: hypergeometric([1], [], -1).is_absolutely_convergent()
                False
                sage: hypergeometric([1], [], -1).n()  # Sum still exists
                0.500000000000000

            Degree `p = q+1` (unit radius of convergence)::

                sage: # needs sage.symbolic
                sage: hypergeometric([2, 3], [4], 6).is_absolutely_convergent()
                False
                sage: hypergeometric([2, 3], [4], 1).is_absolutely_convergent()
                False
                sage: hypergeometric([2, 3], [5], 1).is_absolutely_convergent()
                False
                sage: hypergeometric([2, 3], [6], 1).is_absolutely_convergent()
                True
                sage: hypergeometric([-2, 3], [4],
                ....:                5).is_absolutely_convergent()
                True
                sage: hypergeometric([2, -3], [4],
                ....:                5).is_absolutely_convergent()
                True
                sage: hypergeometric([2, -3], [-4],
                ....:                5).is_absolutely_convergent()
                True
                sage: hypergeometric([2, -3], [-1],
                ....:                5).is_absolutely_convergent()
                False

            Degree giving zero radius of convergence::

                sage: hypergeometric([1, 2, 3], [4],                                    # needs sage.symbolic
                ....:                2).is_absolutely_convergent()
                False
                sage: hypergeometric([1, 2, 3], [4],                                    # needs sage.symbolic
                ....:                1/2).is_absolutely_convergent()
                False
                sage: (hypergeometric([1, 2, -3], [4], 1/2)                             # needs sage.symbolic
                ....:  .is_absolutely_convergent())  # polynomial
                True
            """
            p, q = len(a), len(b)
            if not self.is_termwise_finite():
                return False
            if p <= q:
                return True
            if self.is_terminating():
                return True
            if p == q + 1:
                if abs(z) < 1:
                    return True
                if abs(z) == 1:
                    if real_part(sum(b) - sum(a)) > 0:
                        return True
            return False

        def terms(self, a, b, z, n=None):
            """
            Generate the terms of ``self`` (optionally only ``n`` terms).

            EXAMPLES::

                sage: list(hypergeometric([-2, 1], [3, 4], x).terms())                  # needs sage.symbolic
                [1, -1/6*x, 1/120*x^2]
                sage: list(hypergeometric([-2, 1], [3, 4], x).terms(2))                 # needs sage.symbolic
                [1, -1/6*x]
                sage: list(hypergeometric([-2, 1], [3, 4], x).terms(0))                 # needs sage.symbolic
                []
            """
            if n is None:
                n = Infinity
            t = Integer(1)
            k = 1
            while k <= n:
                yield t
                for aa in a:
                    t *= (aa + k - 1)
                for bb in b:
                    t /= (bb + k - 1)
                t *= z
                if t == 0:
                    break
                t /= k
                k += 1

        def deflated(self, a, b, z):
            r"""
            Rewrite as a linear combination of functions of strictly lower
            degree by eliminating all parameters ``a[i]`` and ``b[j]`` such
            that ``a[i]`` = ``b[i]`` + ``m`` for nonnegative integer ``m``.

            EXAMPLES::

                sage: # needs sage.symbolic
                sage: x = hypergeometric([6, 1], [3, 4, 5], 10)
                sage: y = x.deflated(); y
                1/252*hypergeometric((4,), (7, 8), 10)
                 + 1/12*hypergeometric((3,), (6, 7), 10)
                 + 1/2*hypergeometric((2,), (5, 6), 10)
                 + hypergeometric((1,), (4, 5), 10)
                sage: x.n(); y.n()
                2.87893612686782
                2.87893612686782

                sage: # needs sage.symbolic
                sage: x = hypergeometric([6, 7], [3, 4, 5], 10)
                sage: y = x.deflated(); y
                25/27216*hypergeometric((), (11,), 10)
                 + 25/648*hypergeometric((), (10,), 10)
                 + 265/504*hypergeometric((), (9,), 10)
                 + 181/63*hypergeometric((), (8,), 10)
                 + 19/3*hypergeometric((), (7,), 10)
                 + 5*hypergeometric((), (6,), 10)
                 + hypergeometric((), (5,), 10)
                sage: x.n(); y.n()
                63.0734110716969
                63.0734110716969
            """
            return sum(map(prod, self._deflated()))

        def _deflated(self, a, b, z):
            """
            Private helper to return list of deflated terms.

            EXAMPLES::

                sage: # needs sage.symbolic
                sage: x = hypergeometric([5], [4], 3)
                sage: y = x.deflated(); y
                7/4*hypergeometric((), (), 3)
                sage: x.n(); y.n()
                35.1496896155784
                35.1496896155784
            """
            new = self.eliminate_parameters()
            aa = new.operands()[0].operands()
            bb = new.operands()[1].operands()
            for i, aaa in enumerate(aa):
                for j, bbb in enumerate(bb):
                    m = aaa - bbb
                    if m in ZZ and m > 0:
                        aaaa = aa[:i] + aa[i + 1:]
                        bbbb = bb[:j] + bb[j + 1:]
                        terms = []
                        for k in range(m + 1):
                            # TODO: could rewrite prefactors as recurrence
                            term = binomial(m, k)
                            for c in aaaa:
                                term *= rising_factorial(c, k)
                            for c in bbbb:
                                term /= rising_factorial(c, k)
                            term *= z ** k
                            term /= rising_factorial(aaa - m, k)
                            F = hypergeometric([c + k for c in aaaa],
                                               [c + k for c in bbbb], z)
                            unique = []
                            counts = []
                            for c, f in F._deflated():
                                if f in unique:
                                    counts[unique.index(f)] += c
                                else:
                                    unique.append(f)
                                    counts.append(c)
                            Fterms = zip(counts, unique)
                            terms += [(term * termG, G) for (termG, G) in
                                      Fterms]
                        return terms
            return ((1, new),)


hypergeometric = Hypergeometric()


def closed_form(hyp):
    """
    Try to evaluate ``hyp`` in closed form using elementary
    (and other simple) functions.

    It may be necessary to call :meth:`Hypergeometric.deflated` first to
    find some closed forms.

    EXAMPLES::

        sage: # needs sage.symbolic
        sage: from sage.functions.hypergeometric import closed_form
        sage: var('a b c z')
        (a, b, c, z)
        sage: closed_form(hypergeometric([1], [], 1 + z))
        -1/z
        sage: closed_form(hypergeometric([], [], 1 + z))
        e^(z + 1)
        sage: closed_form(hypergeometric([], [1/2], 4))
        cosh(4)
        sage: closed_form(hypergeometric([], [3/2], 4))
        1/4*sinh(4)
        sage: closed_form(hypergeometric([], [5/2], 4))
        3/16*cosh(4) - 3/64*sinh(4)
        sage: closed_form(hypergeometric([], [-3/2], 4))
        19/3*cosh(4) - 4*sinh(4)
        sage: closed_form(hypergeometric([-3, 1], [var('a')], z))
        -3*z/a + 6*z^2/((a + 1)*a) - 6*z^3/((a + 2)*(a + 1)*a) + 1
        sage: closed_form(hypergeometric([-3, 1/3], [-4], z))
        7/162*z^3 + 1/9*z^2 + 1/4*z + 1
        sage: closed_form(hypergeometric([], [], z))
        e^z
        sage: closed_form(hypergeometric([a], [], z))
        1/((-z + 1)^a)
        sage: closed_form(hypergeometric([1, 1, 2], [1, 1], z))
        (z - 1)^(-2)
        sage: closed_form(hypergeometric([2, 3], [1], x))
        -1/(x - 1)^3 + 3*x/(x - 1)^4
        sage: closed_form(hypergeometric([1/2], [3/2], -5))
        1/10*sqrt(5)*sqrt(pi)*erf(sqrt(5))
        sage: closed_form(hypergeometric([2], [5], 3))
        4
        sage: closed_form(hypergeometric([2], [5], 5))
        48/625*e^5 + 612/625
        sage: closed_form(hypergeometric([1/2, 7/2], [3/2], z))
        1/5*z^2/(-z + 1)^(5/2) + 2/3*z/(-z + 1)^(3/2) + 1/sqrt(-z + 1)
        sage: closed_form(hypergeometric([1/2, 1], [2], z))
        -2*(sqrt(-z + 1) - 1)/z
        sage: closed_form(hypergeometric([1, 1], [2], z))
        -log(-z + 1)/z
        sage: closed_form(hypergeometric([1, 1], [3], z))
        -2*((z - 1)*log(-z + 1)/z - 1)/z
        sage: closed_form(hypergeometric([1, 1, 1], [2, 2], x))
        hypergeometric((1, 1, 1), (2, 2), x)
    """
    if hyp.is_terminating():
        return sum(hyp.terms())

    new = hyp.eliminate_parameters()

    def _closed_form(hyp):
        a, b, z = hyp.operands()
        a, b = a.operands(), b.operands()
        p, q = len(a), len(b)

        if z == 0:
            return Integer(1)
        if p == q == 0:
            return exp(z)
        if p == 1 and q == 0:
            return (1 - z) ** (-a[0])

        if p == 0 and q == 1:
            # TODO: make this require only linear time
            def _0f1(b, z):
                F12 = cosh(2 * sqrt(z))
                F32 = sinh(2 * sqrt(z)) / (2 * sqrt(z))
                if 2 * b == 1:
                    return F12
                if 2 * b == 3:
                    return F32
                if 2 * b > 3:
                    return ((b - 2) * (b - 1) / z * (_0f1(b - 2, z) -
                            _0f1(b - 1, z)))
                if 2 * b < 1:
                    return (_0f1(b + 1, z) + z / (b * (b + 1)) *
                            _0f1(b + 2, z))
                raise ValueError
            # Can evaluate 0F1 in terms of elementary functions when
            # the parameter is a half-integer
            if 2 * b[0] in ZZ and b[0] not in ZZ:
                return _0f1(b[0], z)

        # Confluent hypergeometric function
        if p == 1 and q == 1:
            aa, bb = a[0], b[0]
            if aa * 2 == 1 and bb * 2 == 3:
                t = sqrt(-z)
                return sqrt(pi) / 2 * erf(t) / t
            if a == 1 and b == 2:
                return (exp(z) - 1) / z
            n, m = aa, bb
            if n in ZZ and m in ZZ and m > 0 and n > 0:
                rf = rising_factorial
                if m <= n:
                    return (exp(z) * sum(rf(m - n, k) * (-z) ** k /
                            factorial(k) / rf(m, k) for k in
                            range(n - m + 1)))
                else:
                    T = sum(rf(n - m + 1, k) * z ** k /
                            (factorial(k) * rf(2 - m, k)) for k in
                            range(m - n))
                    U = sum(rf(1 - n, k) * (-z) ** k /
                            (factorial(k) * rf(2 - m, k)) for k in
                            range(n))
                    return (factorial(m - 2) * rf(1 - m, n) *
                            z ** (1 - m) / factorial(n - 1) *
                            (T - exp(z) * U))

        if p == 2 and q == 1:
            R12 = QQ((1, 2))
            R32 = QQ((3, 2))

            def _2f1(a, b, c, z):
                """
                Evaluation of 2F1(a, b; c; z), assuming a, b, c positive
                integers or half-integers
                """
                if b == c:
                    return (1 - z) ** (-a)
                if a == c:
                    return (1 - z) ** (-b)
                if a == 0 or b == 0:
                    return Integer(1)
                if a > b:
                    a, b = b, a
                if b >= 2:
                    F1 = _2f1(a, b - 1, c, z)
                    F2 = _2f1(a, b - 2, c, z)
                    q = (b - 1) * (z - 1)
                    return (((c - 2 * b + 2 + (b - a - 1) * z) * F1 +
                            (b - c - 1) * F2) / q)
                if c > 2:
                    # how to handle this case?
                    if a - c + 1 == 0 or b - c + 1 == 0:
                        raise NotImplementedError
                    F1 = _2f1(a, b, c - 1, z)
                    F2 = _2f1(a, b, c - 2, z)
                    r1 = (c - 1) * (2 - c - (a + b - 2 * c + 3) * z)
                    r2 = (c - 1) * (c - 2) * (1 - z)
                    q = (a - c + 1) * (b - c + 1) * z
                    return (r1 * F1 + r2 * F2) / q

                if (a, b, c) == (R12, 1, 2):
                    return (2 - 2 * sqrt(1 - z)) / z
                if (a, b, c) == (1, 1, 2):
                    return -log(1 - z) / z
                if (a, b, c) == (1, R32, R12):
                    return (1 + z) / (1 - z) ** 2
                if (a, b, c) == (1, R32, 2):
                    return 2 * (1 / sqrt(1 - z) - 1) / z
                if (a, b, c) == (R32, 2, R12):
                    return (1 + 3 * z) / (1 - z) ** 3
                if (a, b, c) == (R32, 2, 1):
                    return (2 + z) / (2 * (sqrt(1 - z) * (1 - z) ** 2))
                if (a, b, c) == (2, 2, 1):
                    return (1 + z) / (1 - z) ** 3
                raise NotImplementedError
            aa, bb = a
            cc, = b
            if z == 1:
                return (gamma(cc) * gamma(cc - aa - bb) / gamma(cc - aa) /
                        gamma(cc - bb))
            if all((cf * 2) in ZZ and cf > 0 for cf in (aa, bb, cc)):
                try:
                    return _2f1(aa, bb, cc, z)
                except NotImplementedError:
                    pass
        return hyp
    return sum([coeff * _closed_form(pfq) for coeff, pfq in new._deflated()])


class Hypergeometric_M(BuiltinFunction):
    r"""
    The confluent hypergeometric function of the first kind,
    `y = M(a,b,z)`, is defined to be the solution to Kummer's differential
    equation

    .. MATH::

        zy'' + (b-z)y' - ay = 0.

    This is not the same as Kummer's `U`-hypergeometric function, though it
    satisfies the same DE that `M` does.

    .. warning::

        In the literature, both are called "Kummer confluent
        hypergeometric" functions.

    EXAMPLES::


        sage: hypergeometric_M(1, 1, 1.)                                                # needs mpmath
        2.71828182845905

        sage: # needs sage.symbolic
        sage: hypergeometric_M(1, 1, 1)
        hypergeometric_M(1, 1, 1)
        sage: hypergeometric_M(1, 1, 1).n(70)                                           # needs mpmath
        2.7182818284590452354
        sage: hypergeometric_M(1, 1, 1).simplify_hypergeometric()
        e
        sage: hypergeometric_M(1, 3/2, 1).simplify_hypergeometric()
        1/2*sqrt(pi)*erf(1)*e
        sage: hypergeometric_M(1, 1/2, x).simplify_hypergeometric()
        (-I*sqrt(pi)*x*erf(I*sqrt(-x))*e^x + sqrt(-x))/sqrt(-x)
    """
    def __init__(self):
        r"""
        TESTS::

            sage: maxima(hypergeometric_M(1,1,x))                                       # needs sage.symbolic
            kummer_m(1,1,_SAGE_VAR_x)
            sage: latex(hypergeometric_M(1,1,x))                                        # needs sage.symbolic
            M\left(1, 1, x\right)
        """
        BuiltinFunction.__init__(self, 'hypergeometric_M', nargs=3,
                                 conversions={'mathematica':
                                              'Hypergeometric1F1',
                                              'maple': 'KummerM',
                                              'maxima': 'kummer_m',
                                              'fricas': 'kummerM'},
                                 latex_name='M')

    def _eval_(self, a, b, z, **kwargs):
        """
        TESTS::

            sage: a, b = var('a,b')                                                     # needs sage.symbolic
            sage: hypergeometric_M(a, b, 0)                                             # needs sage.symbolic
            1
        """
        if not isinstance(z, Expression) and z == 0:
            return Integer(1)
        return

    def _evalf_(self, a, b, z, parent=None, algorithm=None):
        """
        TESTS::

            sage: hypergeometric_M(1,1,1).n()                                           # needs mpmath sage.symbolic
            2.71828182845905
        """
        return _mpmath_utils_call(_mpmath_hyp1f1, a, b, z, parent=parent)

    def _derivative_(self, a, b, z, diff_param):
        """
        TESTS::

            sage: diff(hypergeometric_M(1, 1, x), x, 3)                                 # needs sage.symbolic
            hypergeometric_M(4, 4, x)
            sage: diff(hypergeometric_M(x, 1, 1), x, 3)                                 # needs sage.symbolic
            Traceback (most recent call last):
            ...
            NotImplementedError: derivative of hypergeometric function with respect to parameters
        """
        if diff_param == 2:
            return (a / b) * hypergeometric_M(a + 1, b + 1, z)
        raise NotImplementedError('derivative of hypergeometric function '
                                  'with respect to parameters')

    class EvaluationMethods:
        def generalized(self, a, b, z):
            """
            Return as a generalized hypergeometric function.

            EXAMPLES::

                sage: var('a b z')                                                      # needs sage.symbolic
                (a, b, z)
                sage: hypergeometric_M(a, b, z).generalized()                           # needs sage.symbolic
                hypergeometric((a,), (b,), z)
            """
            return hypergeometric([a], [b], z)


hypergeometric_M = Hypergeometric_M()


class Hypergeometric_U(BuiltinFunction):
    r"""
    The confluent hypergeometric function of the second kind,
    `y = U(a,b,z)`, is defined to be the solution to Kummer's differential
    equation

    .. MATH::

             zy'' + (b-z)y' - ay = 0.

    This satisfies `U(a,b,z) \sim z^{-a}`, as
    `z\rightarrow \infty`, and is sometimes denoted
    `z^{-a}{}_2F_0(a,1+a-b;;-1/z)`. This is not the same as Kummer's
    `M`-hypergeometric function, denoted sometimes as
    `_1F_1(\alpha,\beta,z)`, though it satisfies the same DE that
    `U` does.

    .. warning::

       In the literature, both are called "Kummer confluent
       hypergeometric" functions.

    EXAMPLES::

        sage: # needs mpmath
        sage: hypergeometric_U(1, 1, 1)
        hypergeometric_U(1, 1, 1)
        sage: hypergeometric_U(1, 1, 1.)
        0.596347362323194

        sage: # needs sage.symbolic
        sage: hypergeometric_U(1, 1, 1).n(70)                                           # needs mpmath
        0.59634736232319407434
        sage: hypergeometric_U(10^4, 1/3, 1).n()                                        # needs sage.libs.pari
        6.60377008885811e-35745
        sage: hypergeometric_U(1, 2, 2).simplify_hypergeometric()
        1/2

        sage: hypergeometric_U(2 + I, 2, 1).n()                                         # needs sage.symbolic
        0.183481989942099 - 0.458685959185190*I
        sage: hypergeometric_U(1, 3, x).simplify_hypergeometric()                       # needs sage.symbolic
        (x + 1)/x^2
    """
    def __init__(self):
        r"""
        TESTS::

            sage: maxima(hypergeometric_U(1, 1, x))                                     # needs sage.symbolic
            kummer_u(1,1,_SAGE_VAR_x)
            sage: latex(hypergeometric_U(1, 1, x))                                      # needs sage.symbolic
            U\left(1, 1, x\right)
        """
        BuiltinFunction.__init__(self, 'hypergeometric_U', nargs=3,
                                 conversions={'mathematica':
                                              'HypergeometricU',
                                              'maple': 'KummerU',
                                              'maxima': 'kummer_u',
                                              'fricas': 'kummerU'},
                                 latex_name='U')

    def _eval_(self, a, b, z, **kwargs):
        return

    def _evalf_(self, a, b, z, parent=None, algorithm=None):
        """
        TESTS::

            sage: hypergeometric_U(1, 1, 1).n()                                         # needs mpmath sage.symbolic
            0.596347362323194
        """
        return _mpmath_utils_call(_mpmath_hyperu, a, b, z, parent=parent)

    def _derivative_(self, a, b, z, diff_param):
        """
        TESTS::

            sage: diff(hypergeometric_U(1, 1, x), x, 3)                                 # needs sage.symbolic
            -6*hypergeometric_U(4, 4, x)
            sage: diff(hypergeometric_U(x, 1, 1), x, 3)                                 # needs sage.symbolic
            Traceback (most recent call last):
            ...
            NotImplementedError: derivative of hypergeometric function with respect to parameters
        """
        if diff_param == 2:
            return -a * hypergeometric_U(a + 1, b + 1, z)
        raise NotImplementedError('derivative of hypergeometric function '
                                  'with respect to parameters')

    class EvaluationMethods:
        def generalized(self, a, b, z):
            """
            Return in terms of the generalized hypergeometric function.

            EXAMPLES::

                sage: # needs sage.symbolic
                sage: var('a b z')
                (a, b, z)
                sage: hypergeometric_U(a, b, z).generalized()
                hypergeometric((a, a - b + 1), (), -1/z)/z^a
                sage: hypergeometric_U(1, 3, 1/2).generalized()
                2*hypergeometric((1, -1), (), -2)
                sage: hypergeometric_U(3, I, 2).generalized()
                1/8*hypergeometric((3, -I + 4), (), -1/2)
            """
            return z ** (-a) * hypergeometric([a, a - b + 1], [], -z ** (-1))


hypergeometric_U = Hypergeometric_U()



# sage_setup: distribution = sagemath-categories
r"""
Puiseux Series Ring Element

A Puiseux series is a series of the form

.. MATH::

    p(x) = \sum_{n=N}^{\infty} a_n (x-a)^{n/e},

where the integer :math:`e` is called the *ramification index* of the series
and the number :math:`a` is the *center*. A Puiseux series is essentially a
Laurent series but with fractional exponents.

EXAMPLES:

We begin by constructing the ring of Puiseux series in `x` with coefficients
in the rationals::

    sage: R.<x> = PuiseuxSeriesRing(QQ)

This command also defines ``x`` as the generator of this ring.

When constructing a Puiseux series, the ramification index is automatically
determined from the greatest common divisor of the exponents::

    sage: p = x^(1/2); p
    x^(1/2)
    sage: p.ramification_index()
    2
    sage: q = x^(1/2) + x**(1/3); q
    x^(1/3) + x^(1/2)
    sage: q.ramification_index()
    6

Other arithmetic can be performed with Puiseux Series::

    sage: p + q
    x^(1/3) + 2*x^(1/2)
    sage: p - q
    -x^(1/3)
    sage: p * q
    x^(5/6) + x
    sage: (p / q).add_bigoh(4/3)
    x^(1/6) - x^(1/3) + x^(1/2) - x^(2/3) + x^(5/6) - x + x^(7/6) + O(x^(4/3))

Mind the base ring. However, the base ring can be changed::

    sage: I*q                                                                           # needs sage.rings.number_field
    Traceback (most recent call last):
    ...
    TypeError: unsupported operand parent(s) for *:
    'Number Field in I with defining polynomial x^2 + 1 with I = 1*I' and
    'Puiseux Series Ring in x over Rational Field'
    sage: qz = q.change_ring(ZZ); qz
    x^(1/3) + x^(1/2)
    sage: qz.parent()
    Puiseux Series Ring in x over Integer Ring

Other properties of the Puiseux series can be easily obtained::

    sage: r = (3*x^(-1/5) + 7*x^(2/5) + (1/2)*x).add_bigoh(6/5); r
    3*x^(-1/5) + 7*x^(2/5) + 1/2*x + O(x^(6/5))
    sage: r.valuation()
    -1/5
    sage: r.prec()
    6/5
    sage: r.precision_absolute()
    6/5
    sage: r.precision_relative()
    7/5
    sage: r.exponents()
    [-1/5, 2/5, 1]
    sage: r.coefficients()
    [3, 7, 1/2]

Finally, Puiseux series are compatible with other objects in Sage.
For example, you can perform arithmetic with Laurent series::

    sage: L.<x> = LaurentSeriesRing(ZZ)
    sage: l = 3*x^(-2) + x^(-1) + 2 + x**3
    sage: r + l
    3*x^-2 + x^-1 + 3*x^(-1/5) + 2 + 7*x^(2/5) + 1/2*x + O(x^(6/5))

AUTHORS:

- Chris Swierczewski 2016: initial version on https://github.com/abelfunctions/abelfunctions/tree/master/abelfunctions
- Frédéric Chapoton 2016: integration of code
- Travis Scrimshaw, Sebastian Oehms 2019-2020: basic improvements and completions

REFERENCES:

- :wikipedia:`Puiseux_series`
"""


# ****************************************************************************
#       Copyright (c) 2016 Chris Swierczewski
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.functions import lcm
from sage.arith.misc import gcd
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.rings.infinity import infinity
from sage.rings.laurent_series_ring_element cimport LaurentSeries
from sage.structure.element cimport (Element, AlgebraElement)
from sage.structure.richcmp cimport richcmp


cdef class PuiseuxSeries(AlgebraElement):
    r"""
    A Puiseux series.

    .. MATH::

        \sum_{n=-N}^\infty a_n x^{n/e}

    It is stored as a Laurent series:

    .. MATH::

        \sum_{n=-N}^\infty a_n t^n

    where `t = x^{1/e}`.

    INPUT:

    - ``parent`` -- the parent ring

    - ``f`` -- one of the following types of inputs:

      * instance of :class:`PuiseuxSeries`
      * instance that can be coerced into the Laurent series ring of the parent

    - ``e`` -- integer (default: 1); the ramification index

    EXAMPLES::

        sage: R.<x> = PuiseuxSeriesRing(QQ)
        sage: p = x^(1/2) + x**3; p
        x^(1/2) + x^3
        sage: q = x**(1/2) - x**(-1/2)
        sage: r = q.add_bigoh(7/2); r
        -x^(-1/2) + x^(1/2) + O(x^(7/2))
        sage: r**2
        x^-1 - 2 + x + O(x^3)
    """

    def __init__(self, parent, f, e=1):
        r"""
        Initialize ``self``.

        TESTS::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + x**3
            sage: TestSuite(p).run()
        """
        AlgebraElement.__init__(self, parent)
        L = parent._laurent_series_ring

        if isinstance(f, PuiseuxSeries):
            if (<PuiseuxSeries>f)._l._parent is L:
                l = (<PuiseuxSeries>f)._l
                e = (<PuiseuxSeries>f)._e
            else:
                l = L((<PuiseuxSeries>f)._l)
                e = L((<PuiseuxSeries>f)._e)
        else:
            l = L(f)

        # --------------------------------------------------------
        # choose a representative for this Puiseux series having
        # minimal ramification index. This is necessary because
        # some methods need it as minimal as possible (for example
        # :meth:`laurent_series' or :meth:`power_series`)
        # --------------------------------------------------------
        exp_list = l.exponents()
        prec     = l.prec()
        if prec is infinity:
            d = gcd(exp_list +[e])
        else:
            d = gcd(exp_list + [e] + [prec])
        if d > 1:
            # ramification index can be reduced dividing by d
            e = e / d
            cf_ori = l.list()
            if cf_ori:
                cf = [cf_ori[d*i] for i in range((len(cf_ori)-1) / d + 1)]
            else:
                cf = cf_ori
            val = l.valuation() / d
            l = l.parent()(cf, n=val)
            if prec != infinity:
                l = l.add_bigoh(prec / d)

        self._l = l
        self._e = long(abs(e))

    def __reduce__(self):
        """
        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(1/2) + x**3-x**(-1/4)
            sage: loads(dumps(p)) == p    # indirect doctest
            True
        """
        return (self._parent, (self._l, self._e))

    def _im_gens_(self, codomain, im_gens, base_map=None):
        """
        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(1/3) + x**3
            sage: t = p._im_gens_(QQbar, [2])
            sage: t in QQbar
            True
            sage: f = R.hom([QQbar(2)], check=False)
            sage: t == f(p)
            True
        """
        return self(codomain(im_gens[0]))

    def _repr_(self):
        """
        Return a string representation.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + x**3-x**(-1/4); p
            -x^(-1/4) + x^(1/2) + x^3
            sage: R.zero()
            0

            sage: S.<t> = PuiseuxSeriesRing(Zp(5))                                      # needs sage.rings.padics
            sage: t**(1/2) + 5 * t^(1/3)                                                # needs sage.rings.padics
            (5 + O(5^21))*t^(1/3) + (1 + O(5^20))*t^(1/2)
        """
        laurent = self.laurent_part()
        s = repr(laurent)
        if self.ramification_index() == 1:
            return s

        X = self._parent.variable_name()

        # find a temporary variable name (to avoid multiple transformations)
        Xtemp = '?'
        while Xtemp in s: Xtemp +='?' # if somebody uses '?' in variable_name

        # renaming and generalizing linear term
        s = s.replace('%s' %X, '%s^1' %Xtemp)
        s = s.replace('^1^', '^' )

        # prepare exponent list
        if laurent.prec() is infinity:
            exponents = [ZZ(exp) for exp in set(laurent.exponents())]
        else:
            exponents = [ZZ(exp) for exp in set(laurent.exponents() + [laurent.prec()])]

        # sort exponents such that the largest will be replaced first
        exp_pos = [exp for exp in exponents if exp >= 0]
        exp_pos.sort(reverse=True)
        exp_neg = [exp for exp in exponents if exp < 0]
        exp_neg.sort()
        exponents = exp_neg + exp_pos

        # replacing exponents
        e = ZZ(self.ramification_index())
        for exp_l in exponents:
            exp = exp_l/e
            repl_str = '%s^%s' %(Xtemp, exp_l)
            if exp.is_one():
                s = s.replace(repl_str, '%s' %X)
            elif e.divides(exp_l):
                s = s.replace(repl_str, '%s^%s' %(X, exp))
            else:
                s = s.replace(repl_str, '%s^(%s)' %(X, exp))
        return s

    def __call__(self, x):
        r"""
        Evaluate this Puiseux series.

        INPUT:

        - ``x`` -- element of a ring

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + x**3-x**(-1/4)
            sage: p(16)
            8199/2
            sage: p(pi.n())                                                             # needs sage.symbolic
            32.0276049867404
        """
        # use x.nth_root since x**(1/self._e) returns oo when x = 0
        if isinstance(x, int):
            x = ZZ(x)
        elif isinstance(x, float):
            from sage.rings.complex_mpfr import ComplexField
            x = ComplexField()(x)
        t = x.nth_root(self._e)
        p = self._l.__u.polynomial()
        n = self._l.__n
        return p(t)*t**n

    def _common_ramification_index(self, PuiseuxSeries right):
        r"""
        Return a ramification index common to ``self`` and ``right``.

        In order to perform arithmetic on Puiseux series it is useful to find a
        common ramification index between two operands. That is, given Puiseux
        series :math:`p` and :math:`q` of ramification indices :math:`e` and
        :math:`f` we write both as series :math:`\tilde{f}` and
        :math:`\tilde{g}` in :math:`(x-a)^(1/g)` such that,

        .. MATH::

            f = \tilde{f}((x-a)^M), g = \tilde{g}((x-a)^N).

        INPUT:

        - ``right`` -- a Puiseux series

        OUTPUT:

        - ``g`` -- integer; a ramification index common to ``self`` and
          ``right``
        - ``M``, ``N`` -- integers; scaling factors on ``self`` and ``right``,
          respectively

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/3) + x**2-x**(-1/7)
            sage: q = x^(-1/3) + x**2-x**(1/5)
            sage: p._common_ramification_index(q)
            (105, 5, 7)
            sage: q._common_ramification_index(p)
            (105, 7, 5)
        """
        m = self._e
        n = right._e
        g = lcm(m, n)
        m = g / m
        n = g / n
        return g, m, n

    cpdef _add_(self, right_m):
        """
        Return the sum.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: q = 2*x^(1/3) + 3/4 * x^(2/5)
            sage: p + q                                        # indirect doctest
            2*x^(1/3) + 3/4*x^(2/5) + x^(1/2) + 3/4*x^(2/3)
        """
        cdef PuiseuxSeries right = <PuiseuxSeries>right_m
        cdef LaurentSeries l, l1, l2
        cdef size_t g, m, n

        g, m, n = self._common_ramification_index(right)
        l1 = self._l.V(m)
        l2 = right._l.V(n)
        l = l1 + l2
        return type(self)(self._parent, l, g)

    cpdef _sub_(self, right_m):
        """
        Return the difference.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: q = 2*x^(1/3) + 3/4 * x^(2/5)
            sage: p - q                                        # indirect doctest
            -2*x^(1/3) - 3/4*x^(2/5) + x^(1/2) + 3/4*x^(2/3)
        """
        cdef PuiseuxSeries right = <PuiseuxSeries>right_m
        cdef LaurentSeries l, l1, l2
        cdef size_t g, m, n

        g, m, n = self._common_ramification_index(right)
        l1 = self._l.V(m)
        l2 = right._l.V(n)
        l = l1 - l2
        return type(self)(self._parent, l, g)

    cpdef _mul_(self, right_r):
        """
        Return the product.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: q = 2*x^(1/3) + 3/4 * x^(2/5)
            sage: p * q                                        # indirect doctest
            2*x^(5/6) + 3/4*x^(9/10) + 3/2*x + 9/16*x^(16/15)
        """
        cdef PuiseuxSeries right = <PuiseuxSeries>right_r
        cdef LaurentSeries l, l1, l2
        cdef size_t g, m, n

        g, m, n = self._common_ramification_index(right)
        l1 = self._l.V(m)
        l2 = right._l.V(n)
        l = l1 * l2
        return type(self)(self._parent, l, g)

    cpdef _rmul_(self, Element c):
        """
        Return the right scalar multiplication.

        EXAMPLES::

            sage: P.<y> = PuiseuxSeriesRing(ZZ)
            sage: t = y^(-1/3) + O(y^(0))
            sage: 5*t                                          # indirect doctest
            5*y^(-1/3) + O(1)
        """
        return type(self)(self._parent, self._l._rmul_(c), self._e)

    cpdef _lmul_(self, Element c):
        """
        Return the left scalar multiplication.

        EXAMPLES::

            sage: P.<y> = PuiseuxSeriesRing(Zp(3))                                      # needs sage.rings.padics
            sage: t = y^(2/5) + O(y)                                                    # needs sage.rings.padics
            sage: 5*t                                          # indirect doctest       # needs sage.rings.padics
            (2 + 3 + O(3^20))*y^(2/5) + O(y)
        """
        return type(self)(self._parent, self._l._lmul_(c), self._e)

    cpdef _div_(self, right_r):
        """
        Return the quotient.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: q = 2*x^(1/3) + 3/4 * x^(2/5)
            sage: p / q
            1/2*x^(1/6) - 3/16*x^(7/30) + 9/128*x^(3/10) + 3/8*x^(1/3)
             - 27/1024*x^(11/30) - 9/64*x^(2/5) + 81/8192*x^(13/30)
             + 27/512*x^(7/15) - 243/65536*x^(1/2) - 81/4096*x^(8/15)
             + 729/524288*x^(17/30) + 243/32768*x^(3/5) - 2187/4194304*x^(19/30)
             - 729/262144*x^(2/3) + 6561/33554432*x^(7/10)
             + 2187/2097152*x^(11/15) - 19683/268435456*x^(23/30)
             - 6561/16777216*x^(4/5) + O(x^(5/6))
        """
        cdef PuiseuxSeries right = <PuiseuxSeries>right_r
        cdef LaurentSeries l, l1, l2
        cdef size_t g, m, n

        g, m, n = self._common_ramification_index(right)
        l1 = self._l.V(m)
        l2 = right._l.V(n)
        l = l1 / l2
        return type(self)(self._parent, l, g)

    def __pow__(_self, r, dummy):
        """
        Return the power.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: p ** 3
            x^(3/2) + 9/4*x^(5/3) + 27/16*x^(11/6) + 27/64*x^2
        """
        cdef PuiseuxSeries self = _self
        cdef LaurentSeries l
        cdef size_t e

        r = QQ(r)
        numer = r.numerator()
        denom = r.denominator()

        # if the exponent is integral then do normal exponentiation
        if denom == 1:
            l = self._l ** int(numer)
            e = self._e
        # otherwise, we only exponentiate by a rational number if there is a
        # single term in the Puiseux series
        #
        # (I suppose we could use Taylor series expansions in the general case)
        else:
            if not self.is_monomial():
                raise ValueError('can only exponentiate single term by rational')
            l = self._l.V(numer)
            e = self._e * int(denom)
        return type(self)(self._parent, l, e)

    cpdef _richcmp_(self, right_r, int op):
        r"""
        Comparison of ``self`` and ``right``.

        We say two approximate Puiseux series are equal, if they agree for
        all coefficients up to the *minimum* of the precisions of each.

        Comparison is done in dictionary order going from lowest degree
        to highest degree coefficients with respect to the corresponding
        Laurent series. That means that comparison is performed for
        corresponding `LaurentSeries` instances obtained for the common
        ramification index.

        See :meth:`power_series_ring_element._richcmp_` and
        :meth:`_laurent_series_ring_element._richcmp_` for more
        information.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: q = 2*x^(1/3) + 3/4 * x^(2/5)
            sage: (p < q, p >= q, p == 0, q != 0)
            (True, False, False, True)
            sage: p2 = x^(1/2) + 3/5 * x^(2/3)
            sage: (p2 < p, p2 >= p)
            (True, False)
            sage: p3 = p2.add_bigoh(2/3); p3
            x^(1/2) + O(x^(2/3))
            sage: (p3 == p2, p3 == p, p3 < q, p3 > q)
            (True, True, True, False)
        """
        cdef PuiseuxSeries right = <PuiseuxSeries>right_r
        if self._e == right._e:
            return richcmp(self._l, right._l, op)

        # If both have different ramification indices they must be different as
        # Puiseux series (by the normalization performed in the python constructor).
        # We use the ramification index to order them.
        return richcmp(self._e, right._e, op)

    def __lshift__(self, r):
        """
        Apply :meth:`shift` using the operator `<<`.

        EXAMPLES::

            sage: P.<y> = LaurentPolynomialRing(ZZ)
            sage: R.<x> = PuiseuxSeriesRing(P)
            sage: p = y*x**(-1/3) + 2*y^(-2)*x**(1/2)
            sage: p << 1/3                             # indirect doctest
            y + (2*y^-2)*x^(5/6)
        """
        return self.shift(r)

    def __rshift__(self, r):
        """
        Apply :meth:`shift` with negative argument using the operator `>>`.

        EXAMPLES::

            sage: P.<y> = LaurentPolynomialRing(ZZ)
            sage: R.<x> = PuiseuxSeriesRing(P)
            sage: p = y*x**(-1/3) + 2*y^(-2)*x**(1/2)
            sage: p >> 1/3                             # indirect doctest
            y*x^(-2/3) + (2*y^-2)*x^(1/6)
        """
        return self.shift(-r)

    def __bool__(self):
        """
        Return whether ``self`` is not zero.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: p.is_zero()                     # indirect doctest
            False
            sage: R.zero() != 0
            False
        """
        return bool(self._l)

    def __hash__(self):
        """
        Return a hash of ``self``.

        EXAMPLES::

            sage: from operator import xor
            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: hash(p) == xor(hash(p.laurent_part()), 2)  # indirect doctest
            True
        """
        return hash(self._l) ^ self._e

    def __getitem__(self, r):
        r"""
        Return the coefficient with exponent ``r``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: p[-7/2]
            1
            sage: p[0]
            3
            sage: p[1/2]
            5
            sage: p[3]
            -7
            sage: p[100]
            0
        """
        if isinstance(r, slice):
            start, stop, step = r.start, r.stop, r.step
            n = slice(start * self._e, stop * self._e, step * self._e)
            return PuiseuxSeries(self._parent, self._l[start:stop:step], self._e)
        else:
            n = int(r * self._e)
            return self._l[n]

    def __iter__(self):
        """
        Return an iterator over the coefficients.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: list(p)
            [1, 0, 0, 0, 0, 0, 0, 3, 5, 0, 0, 0, 0, -7]
        """
        return iter(self._l)

    def __copy__(self):
        """
        Since this is immutable, return ``self``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(3/4) + 2*x^(4/5) + 3* x^(5/6)
            sage: p2 = copy(p); p2
            x^(3/4) + 2*x^(4/5) + 3*x^(5/6)
            sage: p == p2
            True
            sage: p is p2
            True
        """
        return self

    def laurent_part(self):
        """
        Return the underlying Laurent series.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: p.laurent_part()
            x^3 + 3/4*x^4
        """
        return self._l

    def ramification_index(self):
        """
        Return the ramification index.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: p.ramification_index()
            6
        """
        return self._e

    def valuation(self):
        r"""
        Return the valuation of ``self``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: p.valuation()
            -7/2

        TESTS::

            sage: R.zero().valuation()
            +Infinity
        """
        return self._l.valuation() / QQ(self._e)

    def add_bigoh(self, prec):
        r"""
        Return the truncated series at chosen precision ``prec``.

        INPUT:

        - ``prec`` -- the precision of the series as a rational number

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: p.add_bigoh(2)
            x^(-7/2) + 3 + 5*x^(1/2) + O(x^2)
            sage: p.add_bigoh(0)
            x^(-7/2) + O(1)
            sage: p.add_bigoh(-1)
            x^(-7/2) + O(x^-1)

        .. NOTE::

            The precision passed to the method is adapted to the common
            ramification index::

                sage: R.<x> = PuiseuxSeriesRing(ZZ)
                sage: p = x**(-1/3) + 2*x**(1/5)
                sage: p.add_bigoh(1/2)
                x^(-1/3) + 2*x^(1/5) + O(x^(7/15))
        """
        if prec is infinity or prec >= self.prec():
            return self

        l_prec = int(prec * self._e)
        l = self._l.add_bigoh(l_prec)
        return PuiseuxSeries(self._parent, l, self._e)

    def change_ring(self, R):
        r"""
        Return ``self`` over a the new ring ``R``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: q = p.change_ring(QQ); q
            x^(-7/2) + 3 + 5*x^(1/2) - 7*x^3
            sage: q.parent()
            Puiseux Series Ring in x over Rational Field
        """
        return self._parent.change_ring(R)(self)

    def is_unit(self):
        r"""
        Return whether ``self`` is a unit.

        A Puiseux series is a unit if and only if its leading coefficient is.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: p.is_unit()
            True
            sage: q = 4 * x^(-7/2) + 3 * x**4
            sage: q.is_unit()
            False
        """
        return self._l.is_unit()

    def is_zero(self):
        """
        Return whether ``self`` is zero.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: p.is_zero()
            False
            sage: R.zero().is_zero()
            True
        """
        return self._l.is_zero()

    def is_monomial(self):
        r"""
        Return whether ``self`` is a monomial.

        This is ``True`` if and only if ``self`` is `x^p` for
        some rational `p`.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(1/2) + 3/4 * x^(2/3)
            sage: p.is_monomial()
            False
            sage: q = x**(11/13)
            sage: q.is_monomial()
            True
            sage: q = 4*x**(11/13)
            sage: q.is_monomial()
            False
        """
        return self._l.is_monomial()

    def list(self):
        r"""
        Return the list of coefficients indexed by the exponents of the
        the corresponding Laurent series.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(3/4) + 2*x^(4/5) + 3* x^(5/6)
            sage: p.list()
            [1, 0, 0, 2, 0, 3]
        """
        return self._l.list()

    def coefficients(self):
        r"""
        Return the list of coefficients.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(3/4) + 2*x^(4/5) + 3* x^(5/6)
            sage: p.coefficients()
            [1, 2, 3]
        """
        return self._l.coefficients()

    def exponents(self):
        r"""
        Return the list of exponents.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x^(3/4) + 2*x^(4/5) + 3* x^(5/6)
            sage: p.exponents()
            [3/4, 4/5, 5/6]
        """
        return [QQ(n) /  self._e for n in self._l.exponents()]

    def __setitem__(self, n, value):
        """
        EXAMPLES::

            sage: R.<t> = PuiseuxSeriesRing(QQ)
            sage: f = t^2 + t^3 + O(t^10)
            sage: f[2] = 5
            Traceback (most recent call last):
            ...
            IndexError: Puiseux series are immutable
        """
        raise IndexError('Puiseux series are immutable')

    def degree(self):
        r"""
        Return the degree of ``self``.

        EXAMPLES::

            sage: P.<y> = PolynomialRing(GF(5))
            sage: R.<x> = PuiseuxSeriesRing(P)
            sage: p = 3*y*x**(-2/3) + 2*y**2*x**(1/5); p
            3*y*x^(-2/3) + 2*y^2*x^(1/5)
            sage: p.degree()
            1/5
        """
        return self._l.degree() / self._e

    def shift(self, r):
        r"""
        Return this Puiseux series multiplied by `x^r`.

        EXAMPLES::

            sage: P.<y> = LaurentPolynomialRing(ZZ)
            sage: R.<x> = PuiseuxSeriesRing(P)
            sage: p = y*x**(-1/3) + 2*y^(-2)*x**(1/2); p
            y*x^(-1/3) + (2*y^-2)*x^(1/2)
            sage: p.shift(3)
            y*x^(8/3) + (2*y^-2)*x^(7/2)
        """
        cdef LaurentSeries l = self._l.shift(r * self._e)
        return PuiseuxSeries(self._parent, l, self._e)

    def truncate(self, r):
        r"""
        Return the Puiseux series of degree `< r`.

        This is equivalent to ``self`` modulo `x^r`.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = (x**(-1/3) + 2*x**3)**2; p
            x^(-2/3) + 4*x^(8/3) + 4*x^6
            sage: q = p.truncate(5); q
            x^(-2/3) + 4*x^(8/3)
            sage: q == p.add_bigoh(5)
            True
        """
        l = self._l.truncate(r * self._e)
        return PuiseuxSeries(self._parent, l, self._e)

    def prec(self):
        r"""
        Return the precision of ``self``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = (x**(-1/3) + 2*x**3)**2; p
            x^(-2/3) + 4*x^(8/3) + 4*x^6
            sage: q = p.add_bigoh(5); q
            x^(-2/3) + 4*x^(8/3) + O(x^5)
            sage: q.prec()
            5
        """
        if self._l.prec() is infinity:
            return infinity
        return self._l.prec() / self._e

    precision_absolute = prec

    def precision_relative(self):
        r"""
        Return the relative precision of the series.

        The relative precision of the Puiseux series is the difference
        between its absolute precision and its valuation.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(GF(3))
            sage: p = (x**(-1/3) + 2*x**3)**2; p
            x^(-2/3) + x^(8/3) + x^6
            sage: q = p.add_bigoh(7); q
            x^(-2/3) + x^(8/3) + x^6 + O(x^7)
            sage: q.precision_relative()
            23/3
        """
        if self.is_zero():
            return 0
        return self.prec() - self.valuation()

    def common_prec(self, p):
        r"""
        Return the minimum precision of `p` and ``self``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = (x**(-1/3) + 2*x**3)**2
            sage: q5 = p.add_bigoh(5); q5
            x^(-2/3) + 4*x^(8/3) + O(x^5)
            sage: q7 = p.add_bigoh(7); q7
            x^(-2/3) + 4*x^(8/3) + 4*x^6 + O(x^7)
            sage: q5.common_prec(q7)
            5
            sage: q7.common_prec(q5)
            5
        """
        if self.prec() is infinity:
            return p.prec()
        elif p.prec() is infinity:
            return self.prec()
        return min(self.prec(), p.prec())

    def variable(self):
        r"""
        Return the variable of ``self``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: p.variable()
            'x'
        """
        return self._parent.variable_name()

    def laurent_series(self):
        r"""
        If ``self`` is a Laurent series, return it as a Laurent series.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(ZZ)
            sage: p = x**(1/2) - x**(-1/2)
            sage: p.laurent_series()
            Traceback (most recent call last):
            ...
            ArithmeticError: self is not a Laurent series
            sage: q = p**2
            sage: q.laurent_series()
            x^-1 - 2 + x
        """
        if self._e != 1:
            raise ArithmeticError('self is not a Laurent series')
        return self._l

    def power_series(self):
        r"""
        If ``self`` is a power series, return it as a power series.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<x> = PuiseuxSeriesRing(QQbar)
            sage: p = x**(3/2) - QQbar(I)*x**(1/2)
            sage: p.power_series()
            Traceback (most recent call last):
            ...
            ArithmeticError: self is not a power series
            sage: q = p**2
            sage: q.power_series()
            -x - 2*I*x^2 + x^3
        """
        try:
            l = self.laurent_series()
            return l.power_series()
        except ArithmeticError:
            raise ArithmeticError('self is not a power series')

    def inverse(self):
        r"""
        Return the inverse of ``self``.

        EXAMPLES::

            sage: R.<x> = PuiseuxSeriesRing(QQ)
            sage: p = x^(-7/2) + 3 + 5*x^(1/2) - 7*x**3
            sage: 1/p
            x^(7/2) - 3*x^7 - 5*x^(15/2) + 7*x^10 + 9*x^(21/2) + 30*x^11 +
            25*x^(23/2) + O(x^(27/2))
        """
        return self.__invert__()



# sage_setup: distribution = sagemath-categories
r"""
Tuples
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from itertools import product, combinations_with_replacement

from sage.arith.misc import binomial
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.rings.integer_ring import ZZ
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


class Tuples(Parent, UniqueRepresentation):
    """
    Return the enumerated set of ordered tuples of S of length k.

    An ordered tuple of length k of set is an ordered selection with
    repetition and is represented by a list of length k containing
    elements of set.

    EXAMPLES::

        sage: S = [1,2]
        sage: Tuples(S,3).list()
        [(1, 1, 1), (2, 1, 1), (1, 2, 1), (2, 2, 1), (1, 1, 2),
         (2, 1, 2), (1, 2, 2), (2, 2, 2)]
        sage: mset = ["s","t","e","i","n"]
        sage: Tuples(mset,2).list()
        [('s', 's'), ('t', 's'), ('e', 's'), ('i', 's'), ('n', 's'),
         ('s', 't'), ('t', 't'), ('e', 't'), ('i', 't'), ('n', 't'),
         ('s', 'e'), ('t', 'e'), ('e', 'e'), ('i', 'e'), ('n', 'e'),
         ('s', 'i'), ('t', 'i'), ('e', 'i'), ('i', 'i'), ('n', 'i'),
         ('s', 'n'), ('t', 'n'), ('e', 'n'), ('i', 'n'), ('n', 'n')]

    ::

        sage: K.<a> = GF(4, 'a')                                                        # needs sage.rings.finite_rings
        sage: mset = sorted((x for x in K if x != 0), key=str)                          # needs sage.rings.finite_rings
        sage: Tuples(mset, 2).list()                                                    # needs sage.rings.finite_rings
        [(1, 1),     (a, 1),     (a + 1, 1),
         (1, a),     (a, a),     (a + 1, a),
         (1, a + 1), (a, a + 1), (a + 1, a + 1)]
    """
    @staticmethod
    def __classcall_private__(cls, S, k):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: T = Tuples(['l','i','t'],2); T
            Tuples of ('l', 'i', 't') of length 2
        """
        return super().__classcall__(cls, tuple(S), k)

    def __init__(self, S, k):
        """
        TESTS::

            sage: T = Tuples([1,2,3],2)
            sage: T == loads(dumps(T))
            True
        """
        self.S = S
        self.k = k
        self._index_list = list({S.index(s) for s in S})
        category = FiniteEnumeratedSets()
        Parent.__init__(self, category=category)

    def __repr__(self):
        """
        TESTS::

            sage: repr(Tuples([1,2,3],2))
            'Tuples of (1, 2, 3) of length 2'
        """
        return "Tuples of %s of length %s" % (self.S, self.k)

    def __iter__(self):
        """
        EXAMPLES::

            sage: S = [1,2]
            sage: Tuples(S,3).list()
            [(1, 1, 1), (2, 1, 1), (1, 2, 1), (2, 2, 1), (1, 1, 2),
             (2, 1, 2), (1, 2, 2), (2, 2, 2)]
            sage: mset = ["s","t","e","i","n"]
            sage: Tuples(mset,2).list()
            [('s', 's'), ('t', 's'), ('e', 's'), ('i', 's'), ('n', 's'),
             ('s', 't'), ('t', 't'), ('e', 't'), ('i', 't'), ('n', 't'),
             ('s', 'e'), ('t', 'e'), ('e', 'e'), ('i', 'e'), ('n', 'e'),
             ('s', 'i'), ('t', 'i'), ('e', 'i'), ('i', 'i'), ('n', 'i'),
             ('s', 'n'), ('t', 'n'), ('e', 'n'), ('i', 'n'), ('n', 'n')]
            sage: Tuples((1,1,2),3).list()
            [(1, 1, 1), (2, 1, 1), (1, 2, 1), (2, 2, 1), (1, 1, 2),
             (2, 1, 2), (1, 2, 2), (2, 2, 2)]
        """
        for p in product(self._index_list, repeat=self.k):
            yield tuple(self.S[i] for i in reversed(p))

    def cardinality(self):
        """
        EXAMPLES::

            sage: S = [1,2,3,4,5]
            sage: Tuples(S,2).cardinality()
            25
            sage: S = [1,1,2,3,4,5]
            sage: Tuples(S,2).cardinality()
            25
        """
        return ZZ(len(self._index_list)).__pow__(self.k)


Tuples_sk = Tuples


class UnorderedTuples(Parent, UniqueRepresentation):
    """
    Return the enumerated set of unordered tuples of S of length k.

    An unordered tuple of length k of set is a unordered selection with
    repetitions of set and is represented by a sorted list of length k
    containing elements from set.

    EXAMPLES::

        sage: S = [1,2]
        sage: UnorderedTuples(S,3).list()
        [(1, 1, 1), (1, 1, 2), (1, 2, 2), (2, 2, 2)]
        sage: UnorderedTuples(["a","b","c"],2).list()
        [('a', 'a'), ('a', 'b'), ('a', 'c'), ('b', 'b'), ('b', 'c'),
         ('c', 'c')]
    """
    @staticmethod
    def __classcall_private__(cls, S, k):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: T = UnorderedTuples(['l','i','t'],2); T
            Unordered tuples of ('l', 'i', 't') of length 2
        """
        return super().__classcall__(cls, tuple(S), k)

    def __init__(self, S, k):
        """
        TESTS::

            sage: T = Tuples([1,2,3],2)
            sage: T == loads(dumps(T))
            True
        """
        self.S = S
        self.k = k
        self._index_list = list({S.index(s) for s in S})
        category = FiniteEnumeratedSets()
        Parent.__init__(self, category=category)

    def __repr__(self):
        """
        TESTS::

            sage: repr(UnorderedTuples([1,2,3],2))
            'Unordered tuples of (1, 2, 3) of length 2'
        """
        return "Unordered tuples of %s of length %s" % (self.S, self.k)

    def __iter__(self):
        """
        EXAMPLES::

            sage: S = [1,2]
            sage: UnorderedTuples(S,3).list()
            [(1, 1, 1), (1, 1, 2), (1, 2, 2), (2, 2, 2)]
            sage: UnorderedTuples(["a","b","c"],2).list()
            [('a', 'a'), ('a', 'b'), ('a', 'c'), ('b', 'b'), ('b', 'c'),
             ('c', 'c')]
            sage: UnorderedTuples([1,1,2],3).list()
            [(1, 1, 1), (1, 1, 2), (1, 2, 2), (2, 2, 2)]
        """
        for ans in combinations_with_replacement(self._index_list, self.k):
            yield tuple(self.S[i] for i in ans)

    def cardinality(self):
        """
        EXAMPLES::

            sage: S = [1,2,3,4,5]
            sage: UnorderedTuples(S,2).cardinality()
            15
        """
        return binomial(len(self._index_list) + self.k - 1, self.k)


UnorderedTuples_sk = UnorderedTuples



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.rings.finite_rings       (because all doctests use finite fields)
"""
Places of function fields: rational
"""

# ****************************************************************************
#       Copyright (C) 2016-2022 Kwankyu Lee <ekwankyu@gmail.com>
#                     2019      Brent Baccala
#                     2021      Jonathan Kliem
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from .place import FunctionFieldPlace


class FunctionFieldPlace_rational(FunctionFieldPlace):
    """
    Places of rational function fields.
    """
    def degree(self):
        """
        Return the degree of the place.

        EXAMPLES::

            sage: F.<x> = FunctionField(GF(2))
            sage: O = F.maximal_order()
            sage: i = O.ideal(x^2 + x + 1)
            sage: p = i.place()
            sage: p.degree()
            2
        """
        if self.is_infinite_place():
            return 1
        else:
            return self._prime.gen().numerator().degree()

    def is_infinite_place(self):
        """
        Return ``True`` if the place is at infinite.

        EXAMPLES::

            sage: F.<x> = FunctionField(GF(2))
            sage: F.places()
            [Place (1/x), Place (x), Place (x + 1)]
            sage: [p.is_infinite_place() for p in F.places()]
            [True, False, False]
        """
        F = self.function_field()
        return self.prime_ideal().ring() == F.maximal_order_infinite()

    def local_uniformizer(self):
        """
        Return a local uniformizer of the place.

        EXAMPLES::

            sage: F.<x> = FunctionField(GF(2))
            sage: F.places()
            [Place (1/x), Place (x), Place (x + 1)]
            sage: [p.local_uniformizer() for p in F.places()]
            [1/x, x, x + 1]
        """
        return self.prime_ideal().gen()

    def residue_field(self, name=None):
        """
        Return the residue field of the place.

        EXAMPLES::

            sage: F.<x> = FunctionField(GF(2))
            sage: O = F.maximal_order()
            sage: p = O.ideal(x^2 + x + 1).place()
            sage: k, fr_k, to_k = p.residue_field()                                     # needs sage.rings.function_field
            sage: k                                                                     # needs sage.rings.function_field
            Finite Field in z2 of size 2^2
            sage: fr_k                                                                  # needs sage.rings.function_field
            Ring morphism:
              From: Finite Field in z2 of size 2^2
              To:   Valuation ring at Place (x^2 + x + 1)
            sage: to_k                                                                  # needs sage.rings.function_field
            Ring morphism:
              From: Valuation ring at Place (x^2 + x + 1)
              To:   Finite Field in z2 of size 2^2
        """
        return self.valuation_ring().residue_field(name=name)

    def _residue_field(self, name=None):
        """
        Return the residue field of the place along with the maps from
        and to it.

        INPUT:

        - ``name`` -- string; name of the generator of the residue field

        EXAMPLES::

            sage: # needs sage.modules
            sage: F.<x> = FunctionField(GF(2))
            sage: O = F.maximal_order()
            sage: i = O.ideal(x^2 + x + 1)
            sage: p = i.place()
            sage: R, fr, to = p._residue_field()
            sage: R
            Finite Field in z2 of size 2^2
            sage: [fr(e) for e in R.list()]
            [0, x, x + 1, 1]
            sage: to(x*(x+1)) == to(x) * to(x+1)
            True
        """
        F = self.function_field()
        prime = self.prime_ideal()

        if self.is_infinite_place():
            K = F.constant_base_field()

            def from_K(e):
                return F(e)

            def to_K(f):
                n = f.numerator()
                d = f.denominator()

                n_deg = n.degree()
                d_deg = d.degree()

                if n_deg < d_deg:
                    return K(0)
                elif n_deg == d_deg:
                    return n.lc() / d.lc()
                else:
                    raise TypeError("not in the valuation ring")
        else:
            O = F.maximal_order()
            K, from_K, _to_K = O._residue_field(prime, name=name)

            def to_K(f):
                if f in O: # f.denominator() is 1
                    return _to_K(f.numerator())
                else:
                    d = F(f.denominator())
                    n = d * f

                    nv = prime.valuation(O.ideal(n))
                    dv = prime.valuation(O.ideal(d))

                    if nv > dv:
                        return K(0)
                    elif dv > nv:
                        raise TypeError("not in the valuation ring")

                    s = ~prime.gen()
                    rd = d * s**dv # in O but not in prime
                    rn = n * s**nv # in O but not in prime
                    return to_K(rn) / to_K(rd)

        return K, from_K, to_K

    def valuation_ring(self):
        """
        Return the valuation ring at the place.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2)); _.<Y> = K[]
            sage: L.<y> = K.extension(Y^2 + Y + x + 1/x)                                # needs sage.rings.function_field
            sage: p = L.places_finite()[0]                                              # needs sage.rings.function_field
            sage: p.valuation_ring()                                                    # needs sage.rings.function_field
            Valuation ring at Place (x, x*y)
        """
        from .valuation_ring import FunctionFieldValuationRing

        return FunctionFieldValuationRing(self.function_field(), self)



# sage_setup: distribution = sagemath-categories
r"""
Iterators over the partitions of an integer

The iterators generate partitions in either increasing or decreasing
lexicographic orders and a partition `P` is represented either in ascending
(`P_i \leq P_{i+1}`) or descending (`P_i \geq P_{i+1}`) orders::

    sage: from sage.combinat.partitions import ZS1_iterator
    sage: for p in ZS1_iterator(4):
    ....:     print(p)
    [4]
    [3, 1]
    [2, 2]
    [2, 1, 1]
    [1, 1, 1, 1]
    sage: from sage.combinat.partitions import AccelDesc_iterator
    sage: for p in AccelDesc_iterator(4):
    ....:     print(p)
    [4]
    [3, 1]
    [2, 2]
    [2, 1, 1]
    [1, 1, 1, 1]
    sage: from sage.combinat.partitions import ZS2_iterator
    sage: for p in ZS2_iterator(4):
    ....:     print(p)
    [1, 1, 1, 1]
    [2, 1, 1]
    [2, 2]
    [3, 1]
    [4]
    sage: from sage.combinat.partitions import AccelAsc_iterator
    sage: for p in AccelAsc_iterator(4):
    ....:     print(p)
    [1, 1, 1, 1]
    [1, 1, 2]
    [1, 3]
    [2, 2]
    [4]

For each of these iterators, this module also provides a ``next`` method that
takes a partition as input and return the next partition in the corresponding
ordering::

    sage: from sage.combinat.partitions import ZS1_next
    sage: ZS1_next([2, 2])
    [2, 1, 1]
    sage: from sage.combinat.partitions import AccelDesc_next
    sage: AccelDesc_next([2, 2])
    [2, 1, 1]
    sage: from sage.combinat.partitions import ZS2_next
    sage: ZS2_next([2, 2])
    [3, 1]
    sage: from sage.combinat.partitions import AccelAsc_next
    sage: AccelAsc_next([2, 2])
    [4]

It is also possible to iterate over the partitions of bounded length::

    sage: from sage.combinat.partitions import ZS1_iterator_nk
    sage: for p in ZS1_iterator_nk(6, 3):
    ....:     print(p)
    [6]
    [5, 1]
    [4, 2]
    [4, 1, 1]
    [3, 3]
    [3, 2, 1]
    [2, 2, 2]

AUTHOR:

- William Stein (2007-07-28): initial version
- Jonathan Bober (2007-07-28): wrote the program ``partitions_c.cc``
  that does all the actual heavy lifting.
- David Coudert (2024-06-01): reshape method :meth:`ZS1_iterator` to ease the
  implementation of :meth:`ZS1_next` and add iterators and next methods based on
  ``ZS2``, ``AccelAsc`` and ``AccelDesc`` from [ZS1998]_ and [KS2012]_.
"""

# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#                     2007 Jonathan Bober <jwbober@gmail.com>
#                     2024 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


cdef inline ZS1_step(list P, int n, int *m, int *h):
    r"""
    Compute the partition following ``P`` in the ordering of the ZS1 algorithm.

    This is a helper method for methods :meth:`ZS1_iterator` and :meth:`ZS1_next`.
    Partition ``P`` is modified in place.

    INPUT:

    - ``P`` -- list of size `n` storing a partition of `n`

    - ``n`` -- integer; the sum of the elements of the partition stored in ``P``

    - ``m`` -- pointer to the index of the last element of the partition

    - ``h`` -- pointer to the last value larger than 1 in ``P``

    .. WARNING::

        The method assumes that input parameters are valid and consistent with
        the ZS1 algorithm as proposed in [ZS1998]_. It modifies the content of
        ``P`` and the values pointed by ``m`` and ``h``.

    EXAMPLES::

        sage: from sage.combinat.partitions import ZS1_iterator
        sage: it = ZS1_iterator(4)
        sage: next(it)
        [4]
        sage: type(_)
        <class 'list'>
    """
    # Invariants
    # (A) P[:m+1] is a partition of n.
    # (B) P[h+1:] is an array of n-(h+1) ones.
    # (C) P[i] > 1 for each i <= h.
    # (D) 0 <= h <= m.
    cdef int r, t
    cdef int mm = m[0]
    cdef int hh = h[0]
    if P[hh] == 2:
        mm += 1
        P[hh] = 1
        hh -= 1
    else:
        t = mm - hh + 1
        r = P[hh] - 1
        P[hh] = r
        while t >= r:
            hh += 1
            P[hh] = r
            t -= r
        if t == 0:
            mm = hh
        else:
            mm = hh + 1
            if t > 1:
                hh += 1
                P[hh] = t
    m[0] = mm
    h[0] = hh


def ZS1_iterator(int n):
    r"""
    Return an iterator over the partitions of ``n``.

    The partitions are generated in the decreasing lexicographic order and each
    partition is represented as a list ``P`` in descending order (i.e.,
    `P_i \geq P_{i+1}`). The method yields lists and not objects of type
    :class:`~sage.combinat.partition.Partition`.

    This is an implementation of the ZS1 algorithm found in [ZS1998]_.

    .. SEEALSO::

        - :meth:`sage.combinat.partitions.ZS2_iterator`
        - :meth:`sage.combinat.partitions.AccelDesc_iterator`

    EXAMPLES::

        sage: from sage.combinat.partitions import ZS1_iterator
        sage: for p in ZS1_iterator(4):
        ....:     print(p)
        [4]
        [3, 1]
        [2, 2]
        [2, 1, 1]
        [1, 1, 1, 1]
        sage: next(ZS1_iterator(4))
        [4]
        sage: type(_)
        <class 'list'>
    """
    # Easy cases.
    if n < 0:
        return
    if not n:
        yield []
        return
    cdef list P = [1] * n
    P[0] = n
    cdef int m = 0
    cdef int h = 0
    yield [n]
    while P[0] != 1:
        ZS1_step(P, n, &m, &h)
        yield P[:m+1]


def ZS1_next(list P):
    r"""
    Return the partition after ``P`` in the ordering of the ZS1 algorithm.

    INPUT:

    - ``P`` -- list encoding a partition of an integer `n` in descending order
      (i.e., `P_i \geq P_{i+1}`)

    EXAMPLES::

        sage: from sage.combinat.partitions import ZS1_iterator, ZS1_next
        sage: P = [4]
        sage: while P:
        ....:     print(P)
        ....:     P = ZS1_next(P)
        [4]
        [3, 1]
        [2, 2]
        [2, 1, 1]
        [1, 1, 1, 1]
        sage: A = [list(p) for p in Partitions(7)]
        sage: all(ZS1_next(p) == q for p, q in zip(A, A[1:]))
        True
    """
    if P[0] == 1:
        return
    if P[-1] != 1:
        return P[:-1] + [P[-1] - 1, 1]

    cdef int n = sum(P)
    cdef int m = len(P) - 1  # index of the last element in P
    # Search for the index h of the last value > 1 in P
    cdef int h = m
    while P[h] == 1:
        h -= 1
    # Let Q be such that Q[:m+1] == P and Q[h+1:] is an array of n-(h+1) ones
    cdef list Q = P + [1] * (n - m - 1)
    ZS1_step(Q, n, &m, &h)
    return Q[:m+1]


def ZS1_iterator_nk(int n, int k):
    r"""
    An iterator for the partitions of `n` of length at most `k` (in the
    decreasing lexicographic order) which returns lists and not objects of type
    :class:`~sage.combinat.partition.Partition`.

    The algorithm is a mild variation on :func:`ZS1_iterator`;
    I would not vow for its speed.

    EXAMPLES::

        sage: from sage.combinat.partitions import ZS1_iterator_nk
        sage: it = ZS1_iterator_nk(4, 3)
        sage: next(it)
        [4]
        sage: type(_)
        <class 'list'>
    """
    # Easy cases.
    if n <= 0:
        if n == 0 and k >= 0:
            yield []
        return
    if k <= 1:
        if k == 1:
            yield [n]
        return
    x = [1]*k
    x[0] = n

    cdef int m = 0
    cdef int h = 0
    cdef int r, t
    yield [n]
    while x[0] != 1:
        # Loop invariants at this point:
        # (A) x[:m+1] is a partition of n.
        # (B) x[h+1:m+1] is an array of m-h ones.
        # (C) x[i] > 1 for each i <= h.
        # (D) 0 <= h <= m < k.
        # Note that x[m+1:] might contain leftover from
        # previous steps; we don't clean up after ourselves.
        if x[h] == 2 and m + 1 < k:
            # We have a 2 in the partition, and the space to
            # spread it into two 1s.
            m += 1
            x[h] = 1
            x[m] = 1
            h -= 1
            yield x[:m+1]
        else:
            t = m - h + 1  # 1 + "the number of 1s to the right of x[h] that belong to the partition"
            r = x[h] - 1

            # This loop finds the largest h such that x[:h] can be completed
            # with integers smaller-or-equal to r=x[h]-1 into a partition of n.
            #
            # We decrement h until it becomes possible.
            while t > (k-h-1) * r:
                # Loop invariants:
                # t = n - sum(x[:h+1]) + 1;
                # r = x[h] - 1; x[h] > 1.
                if h == 0:
                    # No way to make the current partition
                    # lexicographically smaller.
                    return
                h -= 1
                t += r + 1
                r = x[h] - 1
            # Decrement x[h] from r + 1 to r, and replace
            # x[h+1:] by the lexicographically highest array
            # it could possibly be. This means replacing
            # x[h+1:] by the array [r, r, r, ..., r, s],
            # where s is the residue of t modulo r (or
            # nothing if that residue is 0).
            x[h] = r
            while t >= r:
                # Loop invariants: t = n - sum(x[:h+1]) + 1;
                # r = x[h] > 1.
                h += 1
                x[h] = r
                t -= r
            if t == 0:
                m = h
            else:
                m = h + 1
                if t > 1:
                    h += 1
                x[m] = t
            yield x[:m+1]
    # free(x)


cdef inline ZS2_step(list P, int n, int *m, int *h):
    r"""
    Compute the partition following ``P`` in the ordering of the ZS2 algorithm.

    This is a helper method for methods :meth:`ZS2_iterator` and :meth:`ZS2_next`.
    Partition ``P`` is modified in place.

    INPUT:

    - ``P`` -- list of size `n` storing a partition of `n`

    - ``n`` -- integer; the sum of the elements of the partition stored in ``P``

    - ``m`` -- pointer to the index of the last element of the partition

    - ``h`` -- pointer to the last value larger than 1 in ``P``

    .. WARNING::

        The method assumes that input parameters are valid and consistent with
        the ZS2 algorithm as proposed in [ZS1998]_. It modifies the content of
        ``P`` and the values pointed by ``m`` and ``h``.

    EXAMPLES::

        sage: from sage.combinat.partitions import ZS2_iterator
        sage: it = ZS2_iterator(4)
        sage: next(it)
        [1, 1, 1, 1]
        sage: type(_)
        <class 'list'>
    """
    # Invariants
    # (A) P[:m+1] is a partition of n.
    # (B) P[h+1:] is an array of n-(h+1) ones.
    # (C) P[i] > 1 for each i <= h.
    # (D) 0 <= h <= m.
    cdef int mm = m[0]
    cdef int hh = h[0]
    cdef int t, r
    if mm - hh > 1:
        hh += 1
        P[hh] = 2
        mm -= 1
    else:
        t = mm - 2
        while P[t] == P[mm - 1]:
            P[t] = 1
            t -= 1
        hh = t + 1
        P[hh] = P[mm - 1] + 1
        r = P[mm] + P[mm - 1] * (mm - hh - 1)
        P[mm] = 1
        if mm - hh > 1:
            P[mm - 1] = 1
        mm = hh + r - 1
    m[0] = mm
    h[0] = hh


def ZS2_iterator(int n):
    r"""
    Return an iterator over the partitions of ``n``.

    The partitions are generated in the increasing lexicographic order and each
    partition is represented as a list in descending order (i.e., `p_i \geq
    p_{i+1}`).

    This is an implementation of the ZS2 algorithm found in [ZS1998]_.

    .. SEEALSO::

        :meth:`sage.combinat.partitions.ZS1_iterator`

    EXAMPLES::

        sage: from sage.combinat.partitions import ZS2_iterator
        sage: for p in ZS2_iterator(4):
        ....:     print(p)
        [1, 1, 1, 1]
        [2, 1, 1]
        [2, 2]
        [3, 1]
        [4]
        sage: next(ZS2_iterator(4))
        [1, 1, 1, 1]
        sage: type(_)
        <class 'list'>
    """
    # Easy cases.
    if n < 0:
        return
    yield [1] * n
    if n <= 1:
        return
    cdef list P = [1]*n
    P[0] = 2
    cdef int h = 0
    cdef int m = n - 2  # index of the last element of the partition
    yield P[:m+1]
    while P[0] != n:
        ZS2_step(P, n, &m, &h)
        yield P[:m+1]


def ZS2_next(list P):
    r"""
    Return the partition after ``P`` in the ordering of the ZS2 algorithm.

    INPUT:

    - ``P`` -- list encoding a partition of an integer `n` in descending order
      (i.e., `P_i \geq P_{i+1}`)

    EXAMPLES::

        sage: from sage.combinat.partitions import ZS2_iterator, ZS2_next
        sage: P = [1, 1, 1, 1]
        sage: while P:
        ....:     print(P)
        ....:     P = ZS2_next(P)
        [1, 1, 1, 1]
        [2, 1, 1]
        [2, 2]
        [3, 1]
        [4]
    """
    if len(P) <= 1:
        return
    if P[0] == 1:
        return [2] + P[:-2]
    cdef int n = sum(P)
    cdef int m = len(P) - 1  # index of the last element in P
    # Search for the index h of the last value > 1 in P
    cdef int h = m
    while P[h] == 1:
        h -= 1
    # Let Q be such that Q[:m+1] == P and Q[h+1:] is an array of n-(h+1) ones
    cdef list Q = P + [1] * (n - m - 1)
    ZS2_step(Q, n, &m, &h)
    return Q[:m+1]


cdef inline AccelDesc_step(list P, int n, int* m, int* h):
    r"""
    Compute the partition following ``P`` in the ordering of the AccelDesc algorithm.

    This is a helper method for methods :meth:`AccelDesc_iterator` and
    :meth:`AccelDesc_next`. Partition ``P`` is modified in place.

    INPUT:

    - ``P`` -- list of size `n` storing a partition of `n`

    - ``n`` -- integer; the sum of the elements of the partition stored in ``P``

    - ``m`` -- pointer to the index of the last element of the partition

    - ``h`` -- pointer to the last value larger than 1 in ``P``

    .. WARNING::

        The method assumes that input parameters are valid and consistent with
        the AccelDesc algorithm as proposed in [ZS2012]_. It modifies the
        content of ``P`` and the values pointed by ``m`` and ``h``.

    EXAMPLES::

        sage: from sage.combinat.partitions import AccelDesc_iterator
        sage: it = AccelDesc_iterator(4)
        sage: next(it)
        [4]
        sage: type(_)
        <class 'list'>
    """
    # Invariants
    # (A) P[:m+1] is a partition of n.
    # (B) P[h+1:] is an array of n-(h+1) ones.
    # (C) P[i] > 1 for each i <= h.
    # (D) 0 <= h <= m.
    cdef int mm = m[0]
    cdef int hh = h[0]
    cdef int r, t
    if P[hh] == 2:
        mm += 1
        P[hh] = 1
        hh -= 1
    else:
        r = P[hh] - 1
        t = mm - hh + 1
        P[hh] = r
        while t >= r:
            hh += 1
            P[hh] = r
            t -= r
        if not t:
            mm = hh
        else:
            mm = hh + 1
            if t > 1:
                hh += 1
                P[hh] = t
    m[0] = mm
    h[0] = hh


def AccelDesc_iterator(int n):
    r"""
    Return an iterator over the partitions of ``n``.

    The partitions are generated in the increasing lexicographic order and each
    partition is represented as a list in descending order (i.e., `p_i \geq
    p_{i+1}`).

    This is an implementation of the AccelDesc algorithm found in [KS2012]_.

    .. SEEALSO::

        :meth:`sage.combinat.partitions.ZS1_iterator`

    EXAMPLES::

        sage: from sage.combinat.partitions import AccelDesc_iterator
        sage: for p in AccelDesc_iterator(4):
        ....:     print(p)
        [4]
        [3, 1]
        [2, 2]
        [2, 1, 1]
        [1, 1, 1, 1]
        sage: next(AccelDesc_iterator(4))
        [4]
        sage: type(_)
        <class 'list'>

    Check that :meth:`ZS1_iterator` and :meth:`AccelDesc_iterator` generate
    partitions in the same order::

        sage: from sage.combinat.partitions import ZS1_iterator
        sage: from sage.misc.prandom import randint
        sage: n = randint(1, 50)
        sage: all(p == q for p, q in zip(ZS1_iterator(n), AccelDesc_iterator(n)))  # long time
        True
    """
    # Easy cases.
    if n < 0:
        return
    if n <= 1:
        yield [1] * n
        return
    cdef list P = [1] * n
    P[0] = n
    yield [n]
    cdef int m = 0
    cdef int h = 0
    while h >= 0:
        AccelDesc_step(P, n, &m, &h)
        yield P[:m+1]


def AccelDesc_next(list P):
    r"""
    Return the partition after ``P`` in the ordering of the AccelDesc algorithm.

    INPUT:

    - ``P`` -- list encoding a partition of an integer `n` in descending order
      (i.e., `P_i \geq P_{i+1}`)

    EXAMPLES::

        sage: from sage.combinat.partitions import AccelDesc_iterator, AccelDesc_next
        sage: P = [4]
        sage: while P:
        ....:     print(P)
        ....:     P = AccelDesc_next(P)
        [4]
        [3, 1]
        [2, 2]
        [2, 1, 1]
        [1, 1, 1, 1]
        sage: A = [list(p) for p in Partitions(7)]
        sage: all(AccelDesc_next(p) == q for p, q in zip(A, A[1:]))
        True
    """
    if P[0] == 1:
        return
    if P[-1] != 1:
        return P[:-1] + [P[-1] - 1, 1]

    cdef int n = sum(P)
    cdef int m = len(P) - 1  # index of the last element in P
    # Search for the index h of the last value > 1 in P
    cdef int h = m
    while P[h] == 1:
        h -= 1
    # Let Q be such that Q[:m+1] == P and Q[h+1:] is an array of n-(h+1) ones
    cdef list Q = P + [1]*(n - m - 1)
    AccelDesc_step(Q, n, &m, &h)
    return Q[:m+1]


def AccelAsc_iterator(int n):
    r"""
    Return an iterator over the partitions of ``n``.

    The partitions are generated in the increasing lexicographic order and each
    partition is represented as a list in ascending order (i.e., `p_i \leq
    p_{i+1}`).

    This is an implementation of the ``AccelAsc`` algorithm found in [KS2012]_.

    .. SEEALSO::

        :meth:`sage.combinat.partitions.ZS1_iterator`

    EXAMPLES::

        sage: from sage.combinat.partitions import AccelAsc_iterator
        sage: for p in AccelAsc_iterator(4):
        ....:     print(p)
        [1, 1, 1, 1]
        [1, 1, 2]
        [1, 3]
        [2, 2]
        [4]
        sage: next(AccelAsc_iterator(4))
        [1, 1, 1, 1]
        sage: type(_)
        <class 'list'>
    """
    # Easy cases.
    if n < 0:
        return
    if n <= 1:
        yield [1] * n
        return
    cdef list P = [0] * n
    cdef int k = 1
    cdef int y = n - 1
    cdef int x, ell
    while k:
        k -= 1
        x = P[k] + 1
        while 2 * x <= y:
            P[k] = x
            y -= x
            k += 1
        ell = k + 1
        while x <= y:
            P[k] = x
            P[ell] = y
            yield P[:ell + 1]
            x += 1
            y -= 1
        y += x - 1
        P[k] = y + 1
        yield P[:k + 1]


def AccelAsc_next(list P):
    r"""
    Return the partition after ``P`` in the ordering of the ``AccelAsc`` algorithm.

    INPUT:

    - ``P`` -- list encoding a partition of an integer `n` in ascending order
      (i.e., `P_i \leq P_{i+1}`)

    EXAMPLES::

        sage: from sage.combinat.partitions import AccelAsc_next
        sage: P = [1, 1, 1, 1]
        sage: while P:
        ....:     print(P)
        ....:     P = AccelAsc_next(P)
        [1, 1, 1, 1]
        [1, 1, 2]
        [1, 3]
        [2, 2]
        [4]
    """
    if len(P) <= 1:
        return

    cdef int n = sum(P)
    cdef int m = len(P) - 1  # index of the last element in P
    # Let Q be such that Q[:m+1] == P and Q[m+1:] is an array of n-(m+1) ones
    cdef list Q = P + [0]*(n - m - 1)
    cdef int x, y
    cdef int k = m - 1
    x = Q[k] + 1
    y = P[m] - 1
    while 2 * x <= y:
        Q[k] = x
        y -= x
        k += 1
    if x <= y:
        Q[k] = x
        Q[k + 1] = y
        return Q[:k + 2]
    y += x - 1
    Q[k] = y + 1
    return Q[:k + 1]
